# coding: utf-8
"""
    GuinierDiffViewer.py

    Copyright (c) 2018-2019, SAXS-Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from CanvasFrame            import CanvasFrame
from molass_legacy.KekLib.OurMatplotlib          import get_default_colors
from SimpleGuinier          import SimpleGuinier
from molass_legacy._MOLASS.SerialSettings         import get_setting, get_xray_picking

class GuinierDiffFrame( Tk.Frame ):
    def __init__( self, parent, viewer ):
        Tk.Frame.__init__( self, parent)
        self.viewer = viewer

        figsize = ( 15, 5 ) if is_low_resolution() else ( 18, 6 )
        self.canvas_frame = canvas_frame = CanvasFrame( self, figsize=figsize )
        canvas_frame.pack()

        fig = canvas_frame.fig
        ax1  = fig.add_subplot( 131 )
        ax2  = fig.add_subplot( 132 )
        ax3  = fig.add_subplot( 133 )
        fig.tight_layout()
        self.axes = [ax1, ax2, ax3]

    def draw( self, xx, ly1, ly2, zoom=False, colors=None, gcolors=None, guinier_list=None ):
        for ax in self.axes:
            ax.cla()

        for ax in self.axes:
            ax.set_xlabel( 'Q²' )
            ax.set_ylabel( 'Ln(I)' )

        ax1, ax2, ax3 = self.axes

        elultion_no = self.viewer.j.get()
        ax1.set_title( "Guinier Plot of the original %d-th elution" % elultion_no )
        ax2.set_title( "Guinier Plot of the corrected %d-th elution" % elultion_no )
        ax3.set_title( "Rg difference between the original and the corrected" )

        if colors is None:
            color1, color2 = None, None
        else:
            color1, color2 = colors

        if gcolors is None:
            gcolor1, gcolor2 = 'cyan', 'red'
        else:
            gcolor1, gcolor2 = gcolors

        ax1.plot( xx, ly1, 'o', markersize=3, color=color1 )
        ax2.plot( xx, ly2, 'o', markersize=3, color=color2 )

        ax3.plot( xx, ly1, 'o', markersize=3, color=color1 )
        ax3.plot( xx, ly2, 'o', markersize=3, color=color2 )

        if zoom:
            xmin1, xmax1 = ax1.get_xlim()
            ymin1, ymax1 = ax1.get_ylim()
            xmin_   = xmin1 * 0.98 + xmax1 * 0.02
            xmax_   = xmin1 * 0.80 + xmax1 * 0.2
            # ymin_   = ymin1 * 0.3  + ymax1 * 0.7
            # ymax_   = ymin1 * 0.05 + ymax1 * 0.95
            ymin_, ymax_ = -9.0, -3.0
            for ax in self.axes:
                ax.set_xlim( xmin_, xmax_ )
                ax.set_ylim( ymin_, ymax_ )

        if guinier_list is not None:
            guinier1 = guinier_list[0]
            x_  = guinier1.guinier_x
            y_  = guinier1.guinier_y
            for ax in [ax1, ax3]:
                ax.plot( x_, y_, color=gcolor1, marker='o', markersize=8 )

            guinier2 = guinier_list[1]
            x_  = guinier2.guinier_x
            y_  = guinier2.guinier_y
            for ax in [ax2, ax3]:
                ax.plot( x_, y_, color=gcolor2, marker='o', markersize=8 )

            tx  = xmin_*0.6 + xmax_*0.4
            ty  = ymin_*0.2 + ymax_*0.8
            text = [ "Original", "Corrected", "Difference" ]
            for k, ax in enumerate(self.axes):
                ax.text( tx, ty, text[k], alpha=0.2, fontsize=30 )

            ty  = ymin_*0.4 + ymax_*0.6
            ax1.text( tx, ty, "Rg=%.3g" % guinier1.Rg, alpha=0.2, fontsize=30 )
            ax2.text( tx, ty, "Rg=%.3g" % guinier2.Rg, alpha=0.2, fontsize=30 )
            ax3.text( tx, ty, "Rg diff=%.3g" % ( guinier2.Rg - guinier1.Rg ), alpha=0.2, fontsize=30 )

        self.canvas_frame.show()


class ElutionCurveFrame( Tk.Frame ):
    def __init__( self, parent, viewer ):
        Tk.Frame.__init__( self, parent)
        self.viewer = viewer
        self.Q  = get_xray_picking()
        self.x  = np.arange( len(viewer.ecurve_y) )
        self.y  = viewer.ecurve_y

        figsize = ( 10, 2 ) if is_low_resolution() else ( 12, 2 )
        self.canvas_frame = canvas_frame = CanvasFrame( self, figsize=figsize )
        canvas_frame.pack()

        fig = canvas_frame.fig
        ax1  = fig.add_subplot( 111 )
        fig.tight_layout()
        self.ax = ax1
        self.canvas_frame.mpl_canvas.mpl_connect( 'button_press_event', self.move )

    def draw( self, j ):
        ax  = self.ax
        ax.cla()
        ax.set_title( 'Elution Curve at Q=%.3g' % self.Q )
        ax.plot( self.y, color='orange' )
        ymin, ymax = ax.get_ylim()
        ax.set_ylim( ymin, ymax )
        ax.plot( [j, j], [ymin, ymax], color='red' )

        xmin, xmax = ax.get_xlim()
        tx  = xmin * 0.7 + xmax * 0.3
        ty  = ymin * 0.5 + ymax * 0.5
        ax.text( tx, ty, "Double click to move", alpha=0.2, fontsize=30 )

        self.canvas_frame.show()

    def move( self, event ):
        if not event.dblclick:
            return

        if event.inaxes != self.ax or event.xdata is None or event.ydata is None:
            return

        n = np.argmin( np.abs( self.x - event.xdata) + np.abs(self.y - event.ydata) )
        self.viewer.j.set( n )

class GuinierDiffViewer( Tk.Toplevel ):
    def __init__( self, parent, qvector, ecurve_y, data1, data2, j, redraw_on_spin=True ):
        self.parent = parent
        Tk.Toplevel.__init__( self, parent )
        self.title( "GuinierDiffViewer" )

        self.ecurve_y   = ecurve_y
        self.data1  = data1
        self.data2  = data2

        vsize   = len(qvector)//2
        slice_  = slice(0,vsize)
        self.slice_ = slice_

        self.xx  = qvector[slice_]**2

        plot_frame = Tk.Frame( self )
        plot_frame.pack()

        self.gframe1 = GuinierDiffFrame( plot_frame, self )
        self.gframe1.pack()

        lower_frame = Tk.Frame( self )
        lower_frame.pack( fill=Tk.BOTH, expand=Tk.Y )

        self.ecurve_frame = ElutionCurveFrame( lower_frame, self )
        self.ecurve_frame.pack( side=Tk.LEFT )

        panel_frame = Tk.Frame( lower_frame )
        panel_frame.pack( side=Tk.RIGHT, fill=Tk.BOTH, expand=Tk.Y )

        panel_center = Tk.Frame( panel_frame )
        panel_center.pack( side=Tk.BOTTOM, anchor=Tk.CENTER, pady=40 )

        jsize   = data1.shape[0]
        self.j  = Tk.IntVar()
        self.j.set( j )

        redraw_frame = Tk.Frame( panel_center )
        redraw_frame.pack(side=Tk.LEFT, padx=10 )

        label = Tk.Label( redraw_frame, text="Elution No." )
        label.pack( side=Tk.LEFT )

        sinbox = Tk.Spinbox( redraw_frame, textvariable=self.j,
                                            from_=0, to=jsize-1, increment=1, 
                                            justify=Tk.CENTER, width=6, state=Tk.NORMAL )
        sinbox.pack( side=Tk.LEFT, padx=5, pady=10 )

        close_button = Tk.Button( panel_center, text="Close", command=self.close )
        close_button.pack( side=Tk.LEFT, padx=20, pady=10 )

        if redraw_on_spin:
            self.j.trace( 'w', lambda *args: self.draw() )
        else:
            redraw_button = Tk.Button( redraw_frame, text="Redraw", command=self.draw )
            redraw_button.pack( side=Tk.LEFT )

        self.draw()

        X   = parent.winfo_rootx()
        Y   = parent.winfo_rooty()
        print( 'X,Y=', X, Y )

        self.geometry("+%d+%d" % (X+50, Y+50))
        self.update()

    def get_guinier_data( self, j ):

        y1  = self.data1[j,self.slice_,1]
        y2  = self.data2[j,self.slice_,1]

        ly1 = np.log( y1 )
        ly2 = np.log( y2 )

        guinier_list = []
        for data in [ self.data1, self.data2 ]:
            data_ = data[j,:,:]
            sg = SimpleGuinier( data_ )
            guinier_list.append( sg )

        return ly1, ly2, guinier_list

    def draw( self ):
        try:
            j   = self.j.get()
        except:
            return

        colors  = get_default_colors()
        gcolors = [ 'cyan', 'yellow', 'red' ]

        ly1, ly2, guinier_list = self.get_guinier_data( j )

        self.gframe1.draw( self.xx, ly1, ly2, zoom=True, colors=colors[0:2], gcolors=gcolors[0:2], guinier_list=guinier_list[0:2] )
        self.ecurve_frame.draw( j )

    def close( self ):
        self.destroy()
        


"""
    Models.Stochastic.MomentsStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import minimize, basinhopping
from scipy.integrate import quad
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Stochastic.MomentUtils import to_moderate_props
from SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.Stochastic.LognormalPoreFunc import distr_func, lognormal_pore_func
from molass_legacy.Models.Stochastic.LognormalUtils import compute_mu_sigma, compute_mode, compute_mu_sigma_from_mean, compute_mean_from_mode
from molass_legacy.Models.Stochastic.ParamLimits import (
    USE_K, KT_BOUND, N_BOUND,
    MAX_PORESIZE, M1_WEIGHT, M2_WEIGHT, PORESIZE_BOUNDS, PORESIZE_MEAN_BOUND, PORESIZE_STDEV_BOUND)

def moments_study_impl(arg_lrf_src, return_rgs=False, use_unreliables=True, correct_entirely=True, progress_cb=None, show_result=False, debug=False):
    from importlib import reload
    import Selective.DataFilter
    reload(Selective.DataFilter)
    from Selective.DataFilter import DataFilter
    filter = DataFilter(arg_lrf_src)
    lrf_src = filter.get_filtered_src()

    logger = logging.getLogger(__name__)

    x = lrf_src.xr_x
    y = lrf_src.xr_y
    rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs(keep_num_components=False, debug=False)
    peaks = lrf_src.get_peaks()
    egh_moments_list = lrf_src.get_egh_moments_list()
    if use_unreliables:
        from importlib import reload
        import Models.Stochastic.MonoporeMoments as module
        reload(module)
        from molass_legacy.Models.Stochastic.MonoporeMoments import study_monopore_moments_impl
        ret = study_monopore_moments_impl(lrf_src, debug=debug)
        if ret is None:
            return
        mnp_params, temp_rgs, unreliable_indeces, params_scaler = ret
        if len(unreliable_indeces) > 0:
            if correct_entirely:
                corrected_rgs = temp_rgs
            else:
                corrected_rgs = peak_rgs.copy()
                corrected_rgs[unreliable_indeces] = peak_rgs[unreliable_indeces]*0.1 + temp_rgs[unreliable_indeces]*0.9
            logger.info("peak_rgs=%s are replaced by corrected_rgs=%s with unreliable_indeces=%s", peak_rgs, corrected_rgs, unreliable_indeces)
            peak_rgs = corrected_rgs
    else:
        mnp_params = monopore_study(x, y, peaks, peak_rgs, props, egh_moments_list, logger=logger, debug=debug)
        unreliable_indeces = np.array([], dtype=int)
    if mnp_params is None:
        return
    if progress_cb is not None:
        progress_cb(1)
    lnp_params = lnpore_study(x, y, peaks, peak_rgs, props, egh_moments_list, mnp_params, params_scaler, unreliable_indeces, logger=logger, progress_cb=progress_cb, show_result=show_result, debug=debug)
    if progress_cb is not None:
        progress_cb(10)
    if return_rgs:
        return lnp_params, peak_rgs
    else:
        return lnp_params

def monopore_study(x, y, peaks, peak_rgs, props, egh_moments_list, logger=None, progress_cb=None, debug=False):
    num_peaks = len(peaks)
    m_props = to_moderate_props(props)
    if debug:
        print("monopore_study: num_peaks=", num_peaks, "peak_rgs=", peak_rgs, "props=", props, "m_props=", m_props)
    me = 1.5
    mp = 1.5

    use_K = True
    use_basinhopping = True                     # it seems that global optimization is required for this problem
    kt_scale = 100 if use_K else 0.1
    param_scales = np.array([100, kt_scale, 10, 10])    # scales to normalize the parameters for basinhopping

    def mnp_column_objective(p, debug=False):
        if use_basinhopping:
            p *= param_scales
        N_, kt, x0_, poresize_ = p
        if USE_K:
            T_ = kt/N_
        else:
            T_ = kt
        rhov = peak_rgs/poresize_
        rhov[rhov > 1] = 1
        dev_list = []
        for k, (M, rho) in enumerate(zip(egh_moments_list, rhov)):
            M1_ = x0_ + N_ * T_ * (1 - rho)**(me + mp)
            M2_ = 2 * N_ * T_**2 * (1 - rho)**(me + 2*mp)
            dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2)
            if debug:
                print([k], "M1, M1_ = %.3g, %.3g" % (M[0], M1_))
                print([k], "M2, M2_ = %.3g, %.3g" % (M[1], M2_))
        return np.sum(np.asarray(dev_list) * m_props)   # seems better not to weight with props

    if USE_K:
        kt = 500
    else:
        kt = 0.25
    init_params = [2000, kt, 0, 200]
    eghM = egh_moments_list[0]
    x0_upper = eghM[0] - 5*np.sqrt(eghM[1])
    bounds = [N_BOUND, KT_BOUND, (-1000, x0_upper), PORESIZE_BOUNDS]

    if use_basinhopping:
        init_params_ = np.array(init_params)/param_scales
        bounds_ = [(lower/scale, upper/scale) for (lower, upper), scale in zip(bounds, param_scales)]
        res = basinhopping(mnp_column_objective, init_params_, niter=100,
                           minimizer_kwargs=dict(method='Nelder-Mead', bounds=bounds_))
        res.x *= param_scales   # do not forget to scale back the result
    else:
        res = minimize(mnp_column_objective, init_params, bounds=bounds, method='Nelder-Mead')

    if debug:
        print("mnp_column_objective: res.x=", res.x)

    max_m2 = np.max([M[1] for M in egh_moments_list])
    score = np.sqrt(res.fun)/num_peaks/max_m2
    print("max_m2=", max_m2,  "score = ", score)
    """
    Ald     153    0.14
    OA_Ald  155    0.19
    """

    N, kt, x0, poresize = res.x
    if USE_K:
        T = kt/N
    else:
        T = kt
    rhov = peak_rgs/poresize
    rhov[rhov > 1] = 1
    abort = False

    def mnp_scales_objective(p, title=None):
        debug = title is not None
        x0_ = p[0]
        scales = p[1:]
        cy_list = []
        if debug:
            moments = []
        for k, (rho, scale) in enumerate(zip(rhov, scales)):
            ni_ = N * (1 - rho)**me
            ti_ = T * (1 - rho)**mp
            if debug:
                M1_ = x0_ + ni_ * ti_
                M2_ = 2 * ni_ * ti_**2
                moments.append([M1_, M2_])
            cy = scale * robust_single_pore_pdf(x - x0_, ni_, ti_)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        dev = np.sum((ty - y)**2)
        if debug:
            from importlib import reload
            import Models.Stochastic.MonoporeUtils as module
            reload(module)
            from molass_legacy.Models.Stochastic.MonoporeUtils import plot_monopore_moments_state
            mnp_params = N, T, x0_, me, mp, poresize
            ret = plot_monopore_moments_state(title, x, y, peak_rgs, mnp_params, x0_upper, cy_list, ty, moments, egh_moments_list)
            if not ret:
                nonlocal abort
                abort = True
        return dev
    
    area = np.sum(x*y)/np.sum(y)
    scale = 0.5*area*np.max(y)/num_peaks    # task: better to simplify this scale, though it is irrelevant here.
    print("area = ", area)
    init_params = np.concatenate([[x0], props*scale])
    if debug:
        mnp_scales_objective(init_params, title="before MNP minimize")
        if abort:
            return
    bounds = [(-1000, 1000)] + [(0, 10)]*num_peaks
    res = minimize(mnp_scales_objective, init_params, bounds=bounds, method='Nelder-Mead')
    if debug:
        mnp_scales_objective(res.x, title="after MNP minimize")
        if abort:
            return

    x0 = res.x[0]
    scales = res.x[1:]
    return np.concatenate([(N, T, x0, me, mp, poresize), scales])

def lnpore_study(x, y, peaks, peak_rgs, props, egh_moments_list, mnp_params, params_scaler, unreliable_indeces, logger=None, progress_cb=None, show_result=False, debug=False):
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
    num_peaks = len(peaks)
    m_props = to_moderate_props(props)
    logger.info("lnpore_study: num_peaks=%d, mnp_params=%s", num_peaks, mnp_params)
    me = 1.5
    mp = 1.5
    N, T, init_x0, me, mp, poresize = mnp_params[0:6]

    nur = len(unreliable_indeces)
    temp_rgs = peak_rgs.copy()

    def lnp_column_objective(p, debug=False):
        N_, TK_, x0_, mean_, stdev_ = p[0:5]
        if USE_K:
            T_ = TK_/N_
        else:
            T_ = TK_
        mu_, sigma_ = compute_mu_sigma_from_mean(mean_, stdev_)
        if nur > 0:
            temp_rgs[unreliable_indeces] = p[5:]

        dev_list = []
        for k, (M, rg) in enumerate(zip(egh_moments_list, temp_rgs)):
            M1_ = x0_ + N_ * T_    * quad(lambda r : distr_func(r, mu_, sigma_) * (1 - min(1, rg/r))**(me +   mp), rg, MAX_PORESIZE)[0]
            M2_ =   2 * N_ * T_**2 * quad(lambda r : distr_func(r, mu_, sigma_) * (1 - min(1, rg/r))**(me + 2*mp), rg, MAX_PORESIZE)[0]
            dev_list.append(M1_WEIGHT*(M1_ - (M[0]))**2 + M2_WEIGHT*(M2_ - M[1])**2)
            if debug:
                print([k], "N_=%d, T_=%.3g, x0_=%.3g, rg=%.3g" % (N_, T_, x0_, rg))
                print([k], "mean_=%.3g, stdev_=%.3g, mu_=%.3g, sigma_=%.3g" % (mean_, stdev_, mu_, sigma_))
                print([k], "M1, M1_ = %.3g, %.3g" % (M[0], M1_))
                print([k], "M2, M2_ = %.3g, %.3g" % (M[1], M2_))
        return np.sum(np.asarray(dev_list)*m_props)

    stdev = poresize*0.1
    mean = compute_mean_from_mode(poresize, stdev)
    # init_N = min(2000, N)   # too large initial N can be hamful?
    init_N = N
    # init_T = N*T/init_N
    if USE_K:
        init_KT = N*T
    else:
        init_KT = T
    init_params = [init_N, init_KT, init_x0, mean, stdev] + list(temp_rgs[unreliable_indeces])
    if debug or True:
        print("init_params for lnp_column_objective = ", init_params)
        lnp_column_objective(init_params, debug=True)

    mnp_bounds = params_scaler.get_bounds()
    # mnp_bounds[0][1] = 1500     # on what condition?
    bounds = np.vstack([mnp_bounds[0:3], PORESIZE_MEAN_BOUND, PORESIZE_STDEV_BOUND] + [(10, 100)]*nur)

    res = minimize(lnp_column_objective, init_params, bounds=bounds, method='Nelder-Mead')
    if debug or True:
        logger.info("res.x from lnp_column_objective = %s", res.x)
        lnp_column_objective(res.x, debug=True)

    if progress_cb is not None:
        progress_cb(5)

    N, KT, x0, mean, stdev = res.x[0:5]
    if nur > 0:
        temp_rgs[unreliable_indeces] = res.x[5:]
        logger.info("temp_rgs have been updated to %s with unreliable_indeces=%s",  temp_rgs, unreliable_indeces)

    if USE_K:
        T = KT/N
    else:
        T = KT
    mu, sigma = compute_mu_sigma_from_mean(mean, stdev)
    mode = compute_mode(mu, sigma)
    
    logger.info("mode=%.3g from LNP optimized mu=%.3g, sigma=%.3g", mode, mu, sigma)
    logger.info("N for lnp_scales_objective=%d", N)

    abort = False

    def lnp_scales_objective(p, title=None):
        x0_ = p[0]
        scales = p[1:]
        rhov = peak_rgs/poresize
        rhov[rhov > 1] = 1
        cy_list = []
        for k, (rg, scale) in enumerate(zip(temp_rgs, scales)):
            cy = lognormal_pore_func(x, scale, N, T, me, mp, mu, sigma, rg, x0_)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        dev = np.sum((ty - y)**2)
        if title is not None:
            from importlib import reload
            import Models.Stochastic.LognormalUtils as lnp_utils
            reload(lnp_utils)
            from molass_legacy.Models.Stochastic.LnporeUtils import plot_lognormal_fitting_state
            lnp_params = np.concatenate([(N, T, x0, me, mp, mu, sigma), scales])
            ret = plot_lognormal_fitting_state(x, y, lnp_params, peak_rgs, title=title)
            if not ret:
                nonlocal abort
                abort = True
        return dev
    
    init_params = np.concatenate([[x0], mnp_params[6:]])
    logger.info("init_params for lnp_scales_objective=%s", init_params)
    if debug:
        title = "Lognormal PSD Elution Components for %s (before minimize)" % get_in_folder()
        lnp_scales_objective(init_params, title=title)
        if abort:
            print("aborting LNP study")
            return
    bounds = [(x0-100, x0+100)] + [(0, 10)]*num_peaks
    res = minimize(lnp_scales_objective, init_params, bounds=bounds, method='Nelder-Mead')
    if debug or show_result:
        print("showing result of lnp_scales_objective = ", res.x)
        title = "Lognormal PSD Elution Components for %s (after minimize)" % get_in_folder()
        lnp_scales_objective(res.x, title=title)
    x0 = res.x[0]

    return np.concatenate([(N, T, x0, me, mp, mu, sigma), res.x[1:]])


"""
    FlowChangeCandidates.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from sklearn.cluster import KMeans
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy.Elution.CurveUtils import simple_plot

END_FLATTEN_WIDTH = 5
OUTSTANDING_RATIO_LIMIT = 5     # > 3.57 for 20170304, > 4.02 for 20181212, > 4.81 for 20191114_3, > 7.22 for 20190607_1
OK_OUTSTANDING_RATIO = 2.5      # < 2.88 for 20190227_2
STD_RATIO_LIMIT = 3             # < 3.63 for 20190309_3
USE_UV_CORRECTOR = True         # can do without this in 20180602
SIMPLY_RECURSIVE = True

def get_ppk(N, k, gy):
    jj = len(gy) - N
    pp = np.argpartition(np.abs(gy), jj)
    ppN = pp[jj:]

    X = ppN.reshape((N,1))

    kp = k + 2                  # k==3 can group together near, but separate, candidates as in 20201208_3

    kmeans = KMeans(n_clusters=kp, random_state=0).fit(X)
    print("labels=", kmeans.labels_)

    pp3_ = []
    for i in range(kp):
        group = ppN[kmeans.labels_ == i]
        n = np.argmax(abs(gy[group]))
        pp3_.append(group[n])
    pp3 = sorted(pp3_, key=lambda j: -abs(gy[j]))

    return jj, pp, ppN, pp3[0:k].copy()

def get_largest_gradients(y, k, peak_region, return_full_info=False, uv_corrector=None, recursive=True, logger=None, debug=False):
    gy = np.gradient(y)
    gy[0:END_FLATTEN_WIDTH] = 0
    gy[-END_FLATTEN_WIDTH:] = 0
    peak_slice = peak_region.get_slice()
    gy[peak_slice] = 0

    N = 10
    jj, pp, ppN, pp3 = get_ppk(N, k, gy)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        def debug_plot(title_tail=""):
            from DataUtils import get_in_folder
            in_folder = get_in_folder()
            a_curve = peak_region.a_curve
            x = a_curve.x
            with plt.Dp():
                fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
                fig.suptitle("Grouping Debug Plot for " + in_folder + title_tail)
                ax1.set_title("Elution Curve")
                ax2.set_title("Top %d Candidates" % N)
                ax3.set_title("%d Candidates after Grouping" % k)
                simple_plot(ax1, a_curve)
                for ax, ppk in [(ax2, ppN), (ax3, pp3)]:
                    ax.plot(x, gy, label="gradient")
                    for i, p in enumerate(ppk):
                        ax.plot(p, gy[p], "o", label=str(i))
                    ax.legend(loc="lower right")
                fig.tight_layout()
                plt.show()
        debug_plot()

    if return_full_info:
        def compute_outstanding_ratio():
            pp_ = sorted(pp[:jj])
            gy_ = gy[pp_]
            return np.max(np.abs(gy[pp3]))/np.std(gy_)

        outstanding_ratio = compute_outstanding_ratio()

        if outstanding_ratio < OUTSTANDING_RATIO_LIMIT:
            N = 15
            jj, pp, ppN, pp3 = get_ppk(N, k, gy)
            outstanding_ratio_retried = compute_outstanding_ratio()
            logger.info("retried get_ppk with N=%d to possibly improving outstanding_ratio=%.3g from %.3g", N, outstanding_ratio_retried, outstanding_ratio)
            outstanding_ratio = outstanding_ratio_retried
            if debug:
                debug_plot(" (retry)")

        smoothed = False
        corrected_y = None
        if recursive:
            outside = peak_region.get_outside()
            std_ratio = np.std(y[peak_slice])/np.std(y[outside])
            if outstanding_ratio < OUTSTANDING_RATIO_LIMIT or std_ratio > STD_RATIO_LIMIT:
                if USE_UV_CORRECTOR:
                    try:
                        uv_corrector.fit(peak_region)
                        corrected_y = uv_corrector.get_corrected_y()
                        y_for_smoothing = corrected_y
                        logger.info("elution curve for flowchange detection has been corrected due to ratios(%.3g, %.3g).", outstanding_ratio, std_ratio)
                    except:
                        log_exception(logger, "uv_corrector.fit failure: ")
                        y_for_smoothing = y
                else:
                    y_for_smoothing = y
                sy = smooth(y_for_smoothing)
                if SIMPLY_RECURSIVE:
                    sgy, spp3, sratio = get_largest_gradients(sy, k, peak_region, return_full_info=True, recursive=False, logger=logger)[0:3]
                else:
                    # may be deprecated
                    sgy, spp3_extra, sratio = get_largest_gradients(sy, k+3, peak_region, return_full_info=True, recursive=False, logger=logger)[0:3]
                    spp3 = spp3_extra[0:k]
                    spp3 = uv_corrector.get_corrected_ppn(sy, k, spp3_extra, debug=True)
                    pmax = np.max(spp3)
                    pmin = np.min(spp3)
                    pwidth = pmax - pmin
                    ppktop_x = uv_corrector.curve1.primary_peak_x
                    print("spp3[0]=", spp3[0], "ppktop_x=", ppktop_x, "pwidth=", pwidth)
                    if spp3[0] < ppktop_x and pwidth < 10:
                        # as in 20170304
                        pstart = pmax + 10
                        sgy_, spp3_ = get_largest_gradients(sy[pstart:], k, peak_region, recursive=False)
                        spp3[-1] = pstart + spp3_[0]

                if debug:
                    print("ratios = (%.3g, %.3g)" % (outstanding_ratio, sratio))
                    curve1 = uv_corrector.curve1
                    plt.push()
                    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                    fig.suptitle("get_largest_gradients recursive debug")
                    simple_plot(ax1, curve1)
                    ax2.plot(y)
                    for p in pp3:
                        ax2.plot(p, y[p], "^")
                    ax2.plot(sy, color="cyan")
                    for p in spp3:
                        ax2.plot(p, sy[p], "o")
                    fig.tight_layout()
                    plt.show()
                    plt.pop()

                if sratio > OK_OUTSTANDING_RATIO:
                    if logger is not None:
                        logger.info("largest_gradients have been replaced with smoothed ones changing o-ratio from %.3g to %.3g.", outstanding_ratio, sratio)
                    gy, pp3, outstanding_ratio = sgy, spp3, sratio
                    # pp3, outstanding_ratio = spp3, sratio
                    smoothed = True
        return gy, pp3, outstanding_ratio, smoothed, corrected_y
    else:
        return gy, pp3



"""
    Optimizer.FvSynthesizer.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting

NUM_MAJOR_SCORES = get_setting("NUM_MAJOR_SCORES")
# see test_7310_FvScoreSynthesizer.py
VALUE_WEIGHTS7_OLD = [0.35, 0.1, 0.35, 0.1, 0.033, 0.033, 0.034]
VALUE_WEIGHTS7 = [0.25, 0.1, 0.25, 0.1, 0.233, 0.033, 0.034]
VALUE_WEIGHTS8_OLD = [0.35, 0.1, 0.35, 0.1, 0.025, 0.025, 0.025, 0.025]
VALUE_WEIGHTS8 = [0.25, 0.1, 0.25, 0.1, 0.225, 0.025, 0.025, 0.025]

MEAN_WEIGHT = 0.8
STDEV_WEIGHT = 0.2

def synthesize(values, positive_elevate=0):
    value_weights = VALUE_WEIGHTS7 if len(values) == 7 else VALUE_WEIGHTS8
    values = np.asarray(values)+ positive_elevate
    stds = np.std(values, axis=0)
    return MEAN_WEIGHT*np.sqrt(np.sum(value_weights * values**2)) + STDEV_WEIGHT*stds - positive_elevate

def synthesize_demo(values, positive_elevate=0):
    values = np.asarray(values)+ positive_elevate
    stds = np.std(values, axis=0)
    return MEAN_WEIGHT*np.sqrt(np.mean(values**2, axis=0)) + STDEV_WEIGHT*stds - positive_elevate

def trial_demo():
    x = np.linspace(-5, 5, 30)
    y = np.linspace(-5, 5, 30)
    xx, yy = np.meshgrid(x, y)
    zz1 = xx + yy
    zz2 = np.max([xx, yy], axis=0)
    zz3 = np.mean([xx, yy], axis=0) + np.max([xx, yy], axis=0)
    zz4 = synthesize_demo([xx, yy], positive_elevate=5)

    fig, axes = plt.subplots(ncols=4, figsize=(18,6), subplot_kw=dict(projection="3d"))
    fig.suptitle("Comparison of Score Synthesizing Functions", fontsize=20)
    ax1, ax2, ax3, ax4 = axes
    ax1.set_title(r"$z=x+y$")
    ax2.set_title(r"$z=max(x,y)$")
    ax3.set_title(r"$z=mean(x,y)+max(x,y)$")
    ax4.set_title(r"$z=\sqrt{\frac{1}{2}(x^2+y^2)} + std([x,y])$")
    ax1.plot_surface(xx, yy, zz1, cmap="jet")
    ax2.plot_surface(xx, yy, zz2, cmap="jet")
    ax3.plot_surface(xx, yy, zz3, cmap="jet")
    ax4.plot_surface(xx, yy, zz4, cmap="jet")
    zlims = []
    for ax in axes:
        zlims.append(ax.get_zlim())
    zlims = np.array(zlims)
    zmin = np.min(zlims[:,0])
    zmax = np.max(zlims[:,1])
    for ax in axes:
        ax.set_zlim(zmin, zmax)

    fig.tight_layout()
    fig.subplots_adjust(left=0.03, right=0.95, wspace=0.1)
    plt.show()

if __name__ == '__main__':
    # use test_7310_FvScoreSynthesizer.py
    # import seaborn
    # seaborn.set()
    # trial_demo()
    pass



"""
    Theory.RdfCheck.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.stats import linregress
from scipy.integrate import quad
from scipy.interpolate import UnivariateSpline
from scipy.optimize import minimize
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from PlotUtils import align_yaxis_np
from MatrixData import simple_plot_3d
from SvdDenoise import get_denoised_data
from .SolidSphere import phi
from .Rg import compute_corrected_Rg
from Rank.Synthesized import synthesized_data
from Theory.SolidSphere import get_boundary_params, get_boundary_params_simple
from DataUtils import get_in_folder

def demo(parent, sd, pno=0, use_estimated_c=False, use_better_estimate=True, debug=False):
    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[pno]
    f_ = range_[0]
    p_ = range_[1]
    t_ = range_[2]

    f, t = f_, t_
    eslice = slice(f,t)

    Rg, gf, gt, _ = compute_corrected_Rg(sd, ecurve, pno, qv, M[:,eslice], E[:,eslice])
    aslice = slice(0, gt)

    plt.push()
    fig = plt.figure(figsize=(21,7))
    ax1 = fig.add_subplot(131)
    ax2 = fig.add_subplot(132)
    ax3 = fig.add_subplot(133)

    fig.suptitle("Spherical Fitting Demo for " + get_in_folder(), fontsize=30)
    ax1.set_title("LRF Range", fontsize=20)
    ax2.set_title("B(q)/A(q)", fontsize=20)
    ax3.set_title("g(r) - 1", fontsize=20)

    b1, b2, k = get_boundary_params_simple(Rg)
    M_, _ = synthesized_data(qv, M[:,eslice], Rg, cd=2, boundary=b1, k=k)

    i = bisect_right(qv, 0.05)
    c = M[i,:]
    c_ = c[eslice]
    C = np.array([c_, c_**2])
    P = M_ @ np.linalg.pinv(C)

    N = len(qv)

    bq0 = int(N*0.8)
    Bq = P[:,1]
    # Bq -= np.average(Bq[bq0:])

    sq_1 = Bq/P[:,0]

    ax1.plot(c, color='orange')

    ymin, ymax = ax1.get_ylim()
    p = Rectangle(
            (f, ymin),      # (x,y)
            t - f,          # width
            ymax - ymin,    # height
            facecolor   = 'cyan',
            alpha       = 0.2,
        )
    ax1.add_patch(p)

    ax2.plot(qv, sq_1)

    Dmax = 70
    Nr = 200
    dr = Dmax/Nr
    rv = np.linspace(0, Dmax, Nr) + dr/2
    dq = qv[-1]/N
    gr_1 = np.zeros(Nr)

    for k, r in enumerate(rv):
        gr_1[k] = 1/(4*np.pi) * np.sum( sq_1 * np.sinc(qv*r) * qv**2 ) * dq 

    ax3.plot(rv, gr_1)

    sq_1_f = np.zeros(N)
    for k, q in enumerate(qv):
        sq_1_f[k] = 4*np.pi * np.sum( gr_1 * np.sinc(q*rv) * rv**2 ) * dr

    ax2t = ax2.twinx()
    ax2t.plot(qv, sq_1_f)

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()
    plt.pop()



# coding: utf-8
"""
    DecompViewer.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import os
import copy
import numpy                as np
import matplotlib.pyplot    as plt
import matplotlib.patches   as mpl_patches      # 'as patches' does not work properly
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar
from molass_legacy.Models.ElutionCurveModels     import EGH, EGHA, EMG, EMGA
from molass_legacy.ElutionDecomposer         import ElutionDecomposer
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
from DualOptimzer import get_dual_optimized_info, EmgaOptimzer, EghaOptimzer
import matplotlib
from molass_legacy._MOLASS.SerialSettings         import get_setting

RETRY_FIT_ERROR_LIMIT   = 0.0001
ALT_ADOPT_RATIO         = 0.5
USE_DUAL_OPTIMIZER      = True

class DecompViewer( Dialog ):
    def __init__( self, parent, title, sd, mapper, peaks=None, flexible=True, manual=False ):
        self.parent = parent
        self.title_ = title
        self.sd     = sd
        self.mapper = mapper
        self.peaks  = peaks
        self.flexible   = flexible
        self.manual     = manual
        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'

    def show( self ):
        Dialog.__init__( self, self.parent, self.title_ )

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        if self.flexible:
            egh_model   = EGHA()
            emg_model   = EMGA()
        else:
            egh_model   = EGH()
            emg_model   = EMG()

        if True:
            figsize = ( 15, 8 ) if is_low_resolution() else ( 18, 10 )
            fig = plt.figure( figsize=figsize )
            ax1_ = fig.add_subplot( 231 )
            ax2_ = fig.add_subplot( 232 )
            ax3_ = fig.add_subplot( 233 )
            ax4_ = fig.add_subplot( 234 )
            ax5_ = fig.add_subplot( 235 )
            ax6_ = fig.add_subplot( 236 )
            axes_list   = [ [ax1_, ax2_, ax3_], [ax4_, ax5_, ax6_] ]
            model_list  = [ egh_model, emg_model ]
            self.event_axes = [ ax2_, ax5_ ]
        else:
            figsize = ( 15, 5 ) if is_low_resolution() else ( 18, 5 )
            fig = plt.figure( figsize=figsize )
            ax1_ = fig.add_subplot( 131 )
            ax2_ = fig.add_subplot( 132 )
            ax3_ = fig.add_subplot( 133 )
            axes_list   = [ [ax1_, ax2_, ax3_] ]
            model_list  = [ None ]

        self.fig = fig
        self.mpl_canvas = FigureCanvasTkAgg( self.fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        in_folder = get_setting( 'in_folder' )
        title = "Decomposition of Xray elution from " + in_folder
        fig.suptitle( title, fontsize=16 )

        self.draw( axes_list, model_list )
        # print( 'self.fa_list[1].init_fit_recs[0]=', self.fa_list[1].init_fit_recs[0] )

        fig.tight_layout()
        fig.subplots_adjust( top=0.92 )

        if self.mplt_ge_2_2:
            self.mpl_canvas.draw()
        else:
            self.mpl_canvas.show()
        self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
        self.toolbar.update()
        if self.manual:
            self.mpl_canvas.mpl_connect( 'button_press_event', self.manual_decomposer )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.ok_button = w
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_button = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def draw( self, axes_list, model_list ):
        mapper  = self.mapper
        sd      = self.sd

        x_curve = mapper.x_curve
        a_baseline = mapper.a_base + mapper.a_base_adjustment
        x_baseline = mapper.x_base + mapper.x_base_adjustment

        x_corrected = mapper.x_vector - x_baseline
        sd.apply_baseline_correction( mapper.get_mapped_info() )
        data    = sd.intensity_array

        self.y = y = x_corrected

        x   = np.arange( len(y) )

        if self.peaks is None:
            self.peaks = peaks = x_curve.get_emg_peaks()
        else:
            peaks   = self.peaks

        print( 'peaks=', peaks )

        self.chisqr_info_list = []
        self.params_info_list = []
        self.fa_list = []
        last_fa = None
        min_error = None

        for k, model in enumerate(model_list):

            ax1, ax2, ax3 = axes_list[k]
            try:
                fa  = ElutionDecomposer( x_curve, x, y, data, peaks=peaks, retry_valley=True, model=model, deeply=True )
            except:
                etb = ExceptionTracebacker()
                print( etb )
                fa  = last_fa

            fit_error = fa.compute_residual_error()
            print( [k], '------------------------ fit_error=', fit_error )
            if fit_error > RETRY_FIT_ERROR_LIMIT:
                # as of 2018-08-10, this is only for SUB_TRN1
                alt_model = EGHA()
                try:
                    fa_alt  = ElutionDecomposer( x_curve, x, y, data, peaks=peaks, retry_valley=True, model=alt_model, deeply=True )
                except:
                    etb = ExceptionTracebacker()
                    print( etb )
                    fa_alt  = last_fa
                fit_error_alt = fa_alt.compute_residual_error()
                print( [k], '------------------------ fit_error_alt=', fit_error_alt )
                if fit_error_alt < fit_error*ALT_ADOPT_RATIO:
                    fa          = fa_alt
                    fit_error   = fit_error_alt

            if min_error is None or fit_error < min_error:
                min_error = fit_error

            self.fa_list.append( fa )
            last_fa = fa
            name = fa.result_model_name

            ax3_title = None
            if USE_DUAL_OPTIMIZER:
                if name == 'EMGA':
                    optimizer_class = EmgaOptimzer
                elif name == 'EGHA':
                    optimizer_class = EghaOptimzer
                else:
                    optimizer_class = None
                if optimizer_class is None:
                    fit_recs_ax3 = fa.fit_recs
                else:
                    ax3_title = "Optimized simultaneously using %s Model" % name
                    opt_recs = get_dual_optimized_info(optimizer_class, x_curve, x, y, fa.fit_recs )
                    fit_recs_ax3 = opt_recs
            else:
                fit_recs_ax3 = fa.fit_recs

            if ax3_title is None:
                ax3_title = "Decomposition of all peaks using %s Model" % name

            ax1.set_title( "Corrected baseline with LPM and Adjustment" )
            ax2.set_title( "Decomposition of major peaks using %s Model" % name )
            ax3.set_title( ax3_title )

            # ax0.plot( mapper.a_vector )
            # ax0.plot( a_baseline, color='red' )
            ax1.plot( mapper.x_vector, color='orange' )
            ax1.plot( x_baseline, color='red' )
            ax2.plot( x_corrected, color='orange' )
            ax3.plot( x_corrected, color='orange' )

            def draw_decomposition( ax, fit_recs, x, y ):

                x_residual = copy.deepcopy( y )

                for rec in fit_recs:
                    func = rec[1]
                    ey = func( x )
                    ax.plot( ey, label='tau = %g' % func.dict_params['tau'] )
                    x_residual -= ey

                ax.plot( x_residual )
                ax.legend()

            draw_decomposition( ax2, fa.init_fit_recs, x, y )
            draw_decomposition( ax3, fit_recs_ax3, x, y )

            init_chisqrs = [ [ rec[0], rec[2] ]  for rec in fa.init_fit_recs ]
            opt_chisqrs  = [ [ rec[0], rec[2] ]  for rec in fa.fit_recs ]
            opt_params   = [ str( rec[1] ) for rec in fa.fit_recs ]
            self.chisqr_info_list.append( [ init_chisqrs, opt_chisqrs ] )
            self.params_info_list.append( opt_params )

            for ax in axes_list[k]:
                ymin, ymax = ax.get_ylim()
                for peak in peaks:
                    f, t = peak.get_fit_limits()
                    p = mpl_patches.Rectangle(
                            (f, ymin),      # (x,y)
                            t - f,          # width
                            ymax - ymin,    # height
                            facecolor   = 'cyan',
                            alpha       = 0.2,
                        )
                    ax.add_patch( p )

        self.min_error = min_error

    def get_min_error( self ):
        return self.min_error

    def get_chisqr_info_list( self ):
        return self.chisqr_info_list

    def get_params_info_list( self ):
        return self.params_info_list

    def manual_decomposer( self, event ):
        if not event.dblclick:
            return

        if event.inaxes not in self.event_axes or event.xdata is None or event.ydata is None:
            return

        if event.inaxes == self.event_axes[0]:
            i = 0
            model = EGH()
        else:
            i = 1
            model = EMG()

        print( 'manual_decomposer', i )

        from ManualDecomposer import ManualDecomposer

        title_parts = self.title_.split( '-' )

        mdecomp = ManualDecomposer( self, "Manual Decomposer -" + title_parts[-1],  self.y, self.fa_list[i].peak_info, model )
        mdecomp.show()

    def save_the_figure( self, folder, analysis_name ):
        # print( 'save_the_figure: ', folder, analysis_name )
        filename = analysis_name.replace( 'analysis', 'figure' )
        path = os.path.join( folder, filename )
        self.fig.savefig( path )



# coding: utf-8
"""
    BackgroundMaker.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from bisect import bisect_right
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.Tools.OutlineViewer import (format_coord, using_normal_conc,
                OutlineViewerDialog,
                OutlineViewerFrame, OutlineViewer3dFrame, OutlineViewer2dFrame)
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

toggle_btn_texts = ["Show 3D View", "Hide 3D View"]

class BackgroundMakerDialog(OutlineViewerDialog):
    def __init__(self, parent, pre_recog):
        self.logger = logging.getLogger( __name__ )

        self.pre_recog = pre_recog
        self.serial_data = pre_recog.pre_recog_copy
        self.using_normal_conc = using_normal_conc()
        self.using_xray_conc = get_setting('use_xray_conc') == 1
        self.showing_3d_view = 0
        self.applied = False
        self.xray_only = True
        Dialog.__init__(self, parent, "Background Maker", visible=False)
        self.tell_xray_frame_redraw()

    def buttonbox(self):
        bframe = Tk.Frame(self)
        bframe.pack(fill=Tk.X)
        box1 = Tk.Frame(bframe)
        box1.pack(side=Tk.LEFT, anchor=Tk.W)

        self.show_3d_btn = Tk.Button(box1, text=toggle_btn_texts[self.showing_3d_view], command=self.toggle_3d_view)
        self.show_3d_btn.pack(padx=20)

        box2 = Tk.Frame(bframe)
        box2.pack(side=Tk.RIGHT, padx=300)
        super().buttonbox(box2)

    def tell_xray_frame_redraw(self):
        """
        note that elution_canvas.redraw_the_angular_canvas
        must not be called until the construction of this dialog is complete
        when all the xray_frame.canvases have been built.
        """
        xray_frame = self.frames[0]
        elution_canvas = xray_frame.canvases[1]
        self.after(100, elution_canvas.redraw_the_angular_canvas)

    def get_frame_class(self):
        return BackgroundMakerFrame

    def get_bridge_slice(self):
        return self.serial_data.xray_slice

    def toggle_3d_view(self):
        for frame in self.frames:
            if self.showing_3d_view:
                frame.forget_left_frame()
            else:
                frame.pack_left_frame()
        self.showing_3d_view = 1 - self.showing_3d_view
        self.show_3d_btn.config(text=toggle_btn_texts[self.showing_3d_view])

    def apply(self):
        # apply Xray elution canvas
        self.frames[0].canvases[1].apply()
        self.applied = True

    def get_the_background_info(self):
        elution_canvas = self.frames[0].canvases[1]
        angular_canvas = self.frames[0].canvases[2]
        qvector = angular_canvas.x
        intensity = angular_canvas.y
        error_matrix = self.serial_data.intensity_array[angular_canvas.bridge_slice,:,2].T
        N = error_matrix.shape[1]
        error = np.sqrt(np.sum((error_matrix/N)**2, axis=1))    # error propagation rule
        return (np.vstack([qvector, intensity, error]).T, elution_canvas.bridge_slice)

class BackgroundMakerFrame(OutlineViewerFrame):
    def __init__(self, parent, dialog, titles, x, y, data, epos, ecurve, el_color, restrict_list,
                    toggle_info, disable_x_selection=False,
                    averaged_ridge=False,
                    changeable_yscale=False, sub_curve_info=None, pre_recog=None,
                    x_axis_ends=None,
                    bridge_slice=None):
        # print('x.shape=', x.shape, 'y.shape=', y.shape, 'data.shape=', data.shape)

        self.dialog = dialog
        self.data = data
        self.x  = x
        self.y  = y

        self.epos = epos
        self.ecurve = ecurve
        self.el_color = el_color
        self.restrict_list = restrict_list
        self.toggle_info = toggle_info
        self.using_xray_conc = get_setting('use_xray_conc') == 1
        self.using_normal_conc = using_normal_conc()

        Tk.Frame.__init__(self, parent)

        self.build_title()
        self.build_body(titles, disable_x_selection, averaged_ridge, changeable_yscale, sub_curve_info, pre_recog, bridge_slice)
        self.forget_left_frame()
        self.disable_toggle_button()

    def build_title(self):
        label = Tk.Label(self, text="Background for " + get_setting('in_folder'), font=('', 18))
        label.pack(pady=20)

    def get_frame_classes(self):
        return OutlineViewer3dFrame, BackgroundMaker2dFrame

    def forget_3d_canvas(self):
        self.canvases[0].pack_forget()

class BackgroundMaker2dFrame(OutlineViewer2dFrame):
    def __init__(self, parent, parent_obj, figsize, title, x, y, data, color, opos, ocolor, restrict_info,
                sub_curve_info=None, pre_recog=None, ecurve=None, opos_sub=None, disable_selection=False,
                changeable_yscale=False,
                bridge_slice=None,
                toolbar=False, val_label=False):
        self.parent = parent
        self.parent_obj = parent_obj
        self.title = title
        self.x = x
        self.y = y
        self.data = data
        self.color = color
        self.opos = opos
        self.ocolor = ocolor
        self.restrict_info = restrict_info
        self.sub_curve_info = sub_curve_info
        self.pre_recog = pre_recog
        self.serial_data = parent_obj.dialog.serial_data
        self.ecurve = ecurve
        self.showing_uv = pre_recog is not None
        self.opos_sub = opos_sub
        self.changeable_yscale = changeable_yscale
        self.bridge_slice = bridge_slice
        self.toolbar_flag = toolbar
        self.val_label_flag = val_label
        self.max_value = len(y) - 1
        Tk.Frame.__init__(self, parent)

        self.showing_elution = title.find('Elution') >= 0
        self.showing_xray = title.find('Xray') >= 0
        self.normal_plot = self.showing_xray or using_normal_conc() or self.showing_elution

        if self.showing_elution:
            self.disable_selection = False
        else:
            self.disable_selection = True

        self.selection_guide_text = "Drag to suggest where to average"

        self.showing_small_angle = 0
        self.init_xlim = None
        self.set_small_xlim()
        self.cbvar = None

        cframe = Tk.Frame(self)
        cframe.pack()
        self.build_cframe(cframe, figsize)

        bframe = Tk.Frame(self)
        bframe.pack(fill=Tk.X, expand=1)

        if self.toolbar_flag:
            bframe_ = self.add_toolbar(bframe)
        else:
            bframe_ = bframe

        self.build_bframe(bframe_)

    def set_small_xlim(self):
        xmin = -0.01
        xmax = self.x[len(self.x)//4]
        self.small_xlim = (xmin, xmax)

    def get_toggle_xlim(self):
        if self.init_xlim is None:
            self.init_xlim = self.ax.get_xlim()
        return self.small_xlim if self.showing_small_angle else self.init_xlim

    def get_default_elution_slice(self, width=10):
        mtd_elution = self.serial_data.mtd_elution
        if mtd_elution is None:
            start, stop = 0, 10
        else:
            start, stop = mtd_elution.propose_background_range(len(self.x), xy=(self.x, self.y))
        return slice(start, stop)

    def build_bframe(self, bframe):
        if not self.showing_elution:
            label = Tk.Label(bframe, text="")
            label.pack()
            return

        center_frame = Tk.Frame(bframe)
        center_frame.pack(pady=5)

        pick_frame = Tk.Frame(center_frame)
        pick_frame.grid(row=0, column=1)

        for j, t in enumerate(["Elution Ragne", "Num Points"]):
            label = Tk.Label(pick_frame, text=t)
            label.grid(row=0, column=j)

        range_frame = Tk.Frame(pick_frame)
        range_frame.grid(row=1, column=0, padx=5)

        elution_slice = self.get_default_elution_slice()
        self.bridge_slice = elution_slice
        start, stop = elution_slice.start, elution_slice.stop
        self.from_  = Tk.IntVar()
        self.from_.set(self.x[start])
        self.to_  = Tk.IntVar()
        self.to_.set(self.x[stop])
        self.ivars  = [ self.from_, self.to_ ]

        k = -1
        for i, t in enumerate(["from ", "to "]):
            if i == 1:
                k += 1
                space = Tk.Label(range_frame, width=1)
                space.grid(row=0, column=k)

            k += 1
            label = Tk.Label(range_frame, text=t)
            label.grid(row=0, column=k)

            k += 1
            entry   = Tk.Spinbox( range_frame, textvariable=self.ivars[i],
                        from_=0, to=len(self.x)-1, increment=1, 
                        justify=Tk.CENTER, width=6 )
            entry.grid(row=0, column=k)

        self.num_points = Tk.IntVar()
        self.num_points.set(stop - start)
        label = Tk.Label(pick_frame, textvariable=self.num_points)
        label.grid(row=1, column=1)

        self.spinbox_stop_trace = False
        for k, dvar in enumerate(self.ivars):
            dvar.trace('w', lambda *args, k_=k:self.spinbox_tracer(k_))

        self.draw()
        # self.redraw_the_angular_canvas()

    def draw(self, canvas_draw=True):
        super().draw(canvas_draw=False)

        if not self.showing_elution:
            self.ax.set_xlim(self.get_toggle_xlim())

        if canvas_draw:
            self.mpl_canvas.draw()

    def draw_selection(self, ax, restrict_info):
        self.draw_selection_guide(ax)

    def draw_selection_guide(self, ax):
        """
        overriding the super method to adjust to the narrowed range (-0.01, 0.4)
        """
        xmin, xmax = self.get_toggle_xlim()
        ymin, ymax = ax.get_ylim()
        tx = (xmin + xmax)/2
        if self.changeable_yscale and self.log_scale:
            # ty = np.power(10, (-2 + np.log10(ymax))/2 )
            ty = np.power(10, np.log10(ymin)*0.1 + np.log10(ymax)*0.9)
        else:
            ty = (ymin + ymax)/2
        ax.text(tx, ty, self.selection_guide_text, ha="center", va="center", alpha=0.2, fontsize=30)

    def onselect(self, xmin, xmax):
        print('onselect', (xmin, xmax))
        if xmax - xmin < 1e-6:
            return

        self.spinbox_stop_trace = True
        try:
            max_i = len(self.x)-1
            start = min(max_i, bisect_right(self.x, xmin))
            j = min(max_i, bisect_right(self.x, xmax))
            stop = j+1
            self.bridge_slice = slice(start, stop)
            self.from_.set(self.x[start])
            self.to_.set(self.x[j])
            self.num_points.set(stop - start)
            self.update()
            self.draw()
        except:
            if False:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print(etb.last_lines())
        self.spinbox_stop_trace = False
        self.redraw_the_angular_canvas()

    def spinbox_tracer(self, k):
        if self.spinbox_stop_trace:
            return

        values = []
        try:
            for ivar in self.ivars:
                values.append(ivar.get())
        except:
            return

        if values[0] > values[1]:
            print('values=', values)
            if k == 0:
                v = values[1]
            else:
                v = values[0]
            self.spinbox_stop_trace = True
            self.ivars[k].set(v)
            values[k] = v
            self.update()
            self.spinbox_stop_trace = False

        start = values[0]
        stop = values[1]+1
        self.bridge_slice = slice(start, stop)
        self.num_points.set(stop - start)
        self.draw()
        self.update()

    def redraw_the_angular_canvas(self):
        y = np.average(self.data[:, self.bridge_slice], axis=1)

        angular_canvas = self.parent_obj.canvases[2]
        angular_canvas.y = y
        angular_canvas.draw(canvas_draw=False)
        angular_canvas.draw_the_bridge_suggested_curves(self.bridge_slice)
        angular_canvas.mpl_canvas.draw()

    def apply(self):
        # set_setting('xray_bg_pickslice', self.bridge_slice)
        pass

    def draw_the_bridge_suggested_curves(self, bridge_slice):
        ax = self.ax
        x = self.x
        for k in range(bridge_slice.start, bridge_slice.stop):
            y = self.data[:, k]
            ax.plot(x, y, color=self.color, alpha=0.5)



# coding: utf-8
"""

    DataModels.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF

"""

import numpy                as np


def _GuinierPorod( G, Rg, d, x ):
    q1  = 1/Rg * np.sqrt( 3*d/2 )
    D   = G * np.exp( -d/2 ) * np.power( 3*d/2, d/2 ) / np.power( Rg, d )
    y   = np.hstack( [  G * np.exp( - x[ x < q1 ]**2 * Rg**2 / 3 ),
                        D / np.power( x[ x >= q1 ], d )
                     ] )
    return y

class GuinierPorod:
    def __init__( self, G, Rg, d ):
        self.G  = G
        self.Rg = Rg
        self.d  = d

    def __call__( self, x, sigma=0 ):
        # TODO: sigma=0 to be deleted
        return _GuinierPorod( self.G, self.Rg, self.d, x )

class GuinierPorodLmfit:
    def __init__( self, def_func=None ):
        if def_func is None:
            self.def_func   = _GuinierPorod
        else:
            self.def_func   = def_func

    def __call__( self, params, x, y, w  ):
        G   = params['G'].value
        Rg  = params['Rg'].value
        d   = params['d'].value
        y_  = self.def_func( G, Rg, d, x )
        return w * ( y_ - y )
        # return y_ - y

def _GuinierPorodGeneral( G, Rg, d, s, x ):
    if d > s:
        q1  = 1/Rg * np.sqrt( (d - s)*(3 - s)/2 )
        D   = G  / np.power( Rg, d - s ) * np.exp( -(d - s)/2 ) * np.power( (d - s)*(3 - s)/2, (d - s)/2 )
        x_  = x[ x < q1 ]
        y   = np.hstack( [  G /  np.power( x_, s ) * np.exp( - x_**2 * Rg**2 / (3 - s) ),
                            D / np.power( x[ x >= q1 ], d )
                         ] )
    else:
        y   = np.ones( len(x) ) * 9999.9999
    return y

class GuinierPorodGeneral:
    def __init__( self, G, Rg, d, s ):
        self.G  = G
        self.Rg = Rg
        self.d  = d
        self.s  = s

    def __call__( self, x ):
        return _GuinierPorodGeneral( self.G, self.Rg, self.d, self.s, x )

class GuinierPorodGeneralLmfit:
    def __init__( self, def_func=None ):
        if def_func is None:
            self.def_func   = _GuinierPorodGeneral
        else:
            self.def_func   = def_func

    def __call__( self, params, x, y, w  ):
        G   = params['G'].value
        Rg  = params['Rg'].value
        d   = params['d'].value
        s   = params['s'].value
        y_  = self.def_func( G, Rg, d, s, x )
        return w * ( y_ - y )



"""
    EmbedCushion.py

    edit this code while keeping the caller code unchanged

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import os

def embed_cushion(caller):
    try:
        embed_cushion_impl(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None, "embed_cushion_impl: ", n=10)
        raise RuntimeError("embed_cushion_impl")

def embed_cushion_impl_UnifiedDecompResultTest(caller):
    from importlib import reload
    import Decomposer.UnifiedDecompResultTest
    reload(Decomposer.UnifiedDecompResultTest)
    from molass_legacy.Decomposer.UnifiedDecompResultTest import unit_test
    unit_test(caller)

def embed_cushion_impl_OptRecsUtils(caller):
    import copy
    from importlib import reload
    import Decomposer.OptRecsUtils
    reload(Decomposer.OptRecsUtils)
    from molass_legacy.Decomposer.OptRecsUtils import eoii_correct_opt_recs
    frame = caller.editor
    opt_recs_ = copy.deepcopy(frame.opt_recs)
    eoii_correct_opt_recs(frame, opt_recs_, debug=True)

def embed_cushion_impl_(caller):
    from importlib import reload

    import Tools.EmbedCushionUtils
    reload(Tools.EmbedCushionUtils)
    from molass_legacy.Tools.EmbedCushionUtils import get_caller_attr

    import SimTools.ErrorModel
    reload(SimTools.ErrorModel)
    from SimTools.ErrorModel import demo

    gi_in_folder = r"E:\PyTools\Data\20180526\GI"
    if not os.path.exists(gi_in_folder):
        gi_in_folder = r"D:\PyTools\Data\20180526\GI"
    cushion_dict = get_caller_attr(caller, "cushion_dict", {})
    gi_sd = cushion_dict.get("gi_sd")
    if gi_sd is None:
        from molass_legacy.Tools.SdUtils import get_sd
        cushion_dict["gi_sd"] = gi_sd = get_sd(gi_in_folder)

    demo(gi_in_folder, gi_sd)

def embed_cushion_impl_(caller):
    from importlib import reload

    import BoundedLRF.HardSphereDemo
    reload(BoundedLRF.HardSphereDemo)
    from BoundedLRF.HardSphereDemo import demo

    demo()

def embed_cushion_impl_IterativeLrfSolverDemo(caller):
    from importlib import reload

    import Tools.EmbedCushionUtils
    reload(Tools.EmbedCushionUtils)
    from molass_legacy.Tools.EmbedCushionUtils import get_caller_attr

    import Trials.BoundedLRF.IterativeLrfSolverDemo
    reload(Trials.BoundedLRF.IterativeLrfSolverDemo)
    from Trials.BoundedLRF.IterativeLrfSolverDemo import demo, demo2, demo3

    print("embed_cushion")

    pdata, popts = caller.get_preview_data(with_update=False)
    try:
        # demo(caller.dialog, pdata, popts)
        # demo2(caller)
        demo3(caller.dialog, pdata, popts)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "do_devel_test: ", n=10)

def embed_cushion_impl_Recon(caller):
    from importlib import reload
    import BoundedLRF.ReconsiderDemo
    reload(BoundedLRF.ReconsiderDemo)
    from BoundedLRF.ReconsiderDemo import demo

    try:
        demo(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "do_devel_test: ", n=10)

def embed_cushion_impl_IFT(caller):
    from importlib import reload
    import BoundedLRF.IftDemo
    reload(BoundedLRF.IftDemo)
    from BoundedLRF.IftDemo import demo

    try:
        demo(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "do_devel_test: ", n=10)

def embed_cushion_impl_ErrorMonteCarlo(caller):
    from importlib import reload
    import BoundedLRF.ErrorMonteCarlo
    reload(BoundedLRF.ErrorMonteCarlo)
    from BoundedLRF.ErrorMonteCarlo import demo

    try:
        demo(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "do_devel_test: ", n=10)

def embed_cushion_impl_Error(caller):
    from importlib import reload
    import BoundedLRF.ErrorCorrectionDemo
    reload(BoundedLRF.ErrorCorrectionDemo)
    from BoundedLRF.ErrorCorrectionDemo import demo

    try:
        demo(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "do_devel_test: ", n=10)

def embed_cushion_impl_EdmParamInspect(caller):
    from importlib import reload
    import Models.RateTheory.EdmParamInspect
    reload(Models.RateTheory.EdmParamInspect)
    from molass_legacy.Models.RateTheory.EdmParamInspect import demo

    try:
        demo(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "do_devel_test: ", n=10)

def embed_cushion_impl_DecompDummy(caller):
    from importlib import reload
    import RangeEditors.DecompDummyDialog
    reload(RangeEditors.DecompDummyDialog)
    from RangeEditors.DecompDummyDialog import decompose

    try:
        decompose(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "decompose: ", n=10)

def embed_cushion_impl_Models(caller):
    from importlib import reload
    import Models.ElutionModelDemo
    reload(Models.ElutionModelDemo)
    from molass_legacy.Models.ElutionModelDemo import demo

    try:
        demo(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "do_devel_test: ", n=10)

def embed_cushion_impl_DispersiveUtils(caller):
    from importlib import reload
    import Models.Stochastic.DispersiveUtils
    reload(Models.Stochastic.DispersiveUtils)
    from molass_legacy.Models.Stochastic.DispersiveUtils import investigate_sdm_params_from_v2_result

    try:
        investigate_sdm_params_from_v2_result(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.dialog.logger, "investigate_sdm_params_from_v2_result: ", n=10)

def embed_cushion_impl_SdmEstimator(caller):
    from importlib import reload
    import Estimators.SdmEstimator
    reload(Estimators.SdmEstimator)
    from Estimators.SdmEstimator import onthefly_test

    try:
        onthefly_test(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "onthefly_test: ", n=10)

def embed_cushion_impl_ParamsIterator(caller):
    from importlib import reload
    import Optimizer.ParamsIterator
    reload(Optimizer.ParamsIterator)
    from molass_legacy.Optimizer.ParamsIterator import iterator_test_from_dialog

    try:
        iterator_test_from_dialog(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "iterator_test_from_dialog: ", n=10)

def embed_cushion_impl_EdmEstimator(caller):
    from importlib import reload
    import Estimators.EdmEstimator
    reload(Estimators.EdmEstimator)
    from Estimators.EdmEstimator import onthefly_test

    try:
        onthefly_test(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "EdmEstimator onthefly_test: ", n=10)

def embed_cushion_impl_RatioInterpretIllust(caller):
    from importlib import reload
    import Optimizer.RatioInterpretIllust
    reload(Optimizer.RatioInterpretIllust)
    from molass_legacy.Optimizer.RatioInterpretIllust import ratio_interpret_illust

    try:
        ratio_interpret_illust(caller.canvas)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "ratio_interpret_illust: ", n=10)

def ScaleAdjustInspect_impl(caller):
    from importlib import reload
    import Optimizer.ScaleAdjustInspect
    reload(Optimizer.ScaleAdjustInspect)
    from molass_legacy.Optimizer.ScaleAdjustInspect import scale_adjust_inspect

    try:
        scale_adjust_inspect(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "scale_adjust_inspect: ", n=10)

def RgCurveInspect_impl(caller):
    from importlib import reload
    import Optimizer.RgCurveInspect
    reload(Optimizer.RgCurveInspect)
    from molass_legacy.Optimizer.RgCurveInspect import rg_curve_inspect

    try:
        rg_curve_inspect(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "rg_curve_inspect: ", n=10)

def RestartPatcher_impl(caller):
    from importlib import reload
    import Optimizer.RestartPatcher
    reload(Optimizer.RestartPatcher)
    from molass_legacy.Optimizer.RestartPatcher import patch_and_restart

    try:
        patch_and_restart(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "patch_and_restart_test: ", n=10)

def BoundsInspect_impl(caller):
    from importlib import reload
    import Optimizer.Devel.BoundsInspect
    reload(Optimizer.Devel.BoundsInspect)
    from molass_legacy.Optimizer.Devel.BoundsInspect import bounds_inspect_impl

    try:
        bounds_inspect_impl(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "bounds_inspect_impl: ", n=10)

def McmcTrial_impl(caller):
    from importlib import reload
    import Solvers.MCMC.TrialBridge
    reload(Solvers.MCMC.TrialBridge)
    from Solvers.MCMC.TrialBridge import mcmc_trial_impl

    try:
        mcmc_trial_impl(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "mcmc_trial_impl: ", n=10)

def SmcTrial_impl(caller):
    from importlib import reload
    import Solvers.SMC.TrialBridge
    reload(Solvers.SMC.TrialBridge)
    from Solvers.SMC.TrialBridge import smc_trial_impl

    try:
        smc_trial_impl(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "smc_trial_impl: ", n=10)

def AbcTrial_impl(caller):
    from importlib import reload
    import Solvers.ABC.TrialBridge
    reload(Solvers.ABC.TrialBridge)
    from Solvers.ABC.TrialBridge import abc_trial_impl

    try:
        abc_trial_impl(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "abc_trial_impl: ", n=10)

def NestTrial_impl(caller):
    from importlib import reload
    import Solvers.UltraNest.TrialBridge
    reload(Solvers.UltraNest.TrialBridge)
    from Solvers.UltraNest.TrialBridge import nest_trial_impl

    try:
        nest_trial_impl(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "nest_trial_impl: ", n=10)

def Plot_callback_txt(caller):
    from importlib import reload
    import Solvers.UltraNest.CallbacktxtPlot
    reload(Solvers.UltraNest.CallbacktxtPlot)
    from Solvers.UltraNest.CallbacktxtPlot import plot_callback_txt_impl

    try:
        plot_callback_txt_impl(caller)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(caller.logger, "plot_callback_txt_impl: ", n=10)

def debug_objective_function(caller):
    from importlib import reload
    import Optimizer.SimpleDebugUtils
    reload(Optimizer.SimpleDebugUtils)
    from molass_legacy.Optimizer.SimpleDebugUtils import debug_optimizer
    debug_optimizer(caller.fullopt)

def embed_cushion_impl(caller):
    import molass_legacy.KekLib.DebugPlot as plt

    extra_button_specs = [
        ("ScaleAdjust Inspect", lambda: ScaleAdjustInspect_impl(caller)),
        ("RgCurve Inspect", lambda: RgCurveInspect_impl(caller)),
        ("Restart Patcher", lambda: RestartPatcher_impl(caller)),
        ("Bounds Inspect", lambda: BoundsInspect_impl(caller)),
        ("MCMC Trial", lambda: McmcTrial_impl(caller)),
        ("SMC Trial", lambda: SmcTrial_impl(caller)),
        ("ABC Trial", lambda: AbcTrial_impl(caller)),
        ("Nest Trial", lambda: NestTrial_impl(caller)),
        ("Plot callback.txt", lambda: Plot_callback_txt(caller)),
        ("Objective Function Debug", lambda: debug_objective_function(caller)),
    ]

    with plt.Dp(button_spec=["OK", "Cancel"], extra_button_specs=extra_button_specs):
        fig, ax = plt.subplots()
        plt.show()


# coding: utf-8
"""
    DualEghMixture.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import copy
import numpy as np
import scipy.stats as stats
from scipy.optimize import minimize
from sklearn.cluster import KMeans
import molass_legacy.KekLib.DebugPlot as plt
from .EghMixture import egh_pdf, e1, e2, e3, EghMixture

USE_MODE_AS_TR_INIT = True
USE_MODERATE_SIGMAS = True
MODERATE_SIGMAS_RATIO = 1.5
APPLY_PARAM_CONSTRAINTS = True

class DualEghMixture:
    def __init__(self, K, max_iter=100, random_states=None, anim_data=False):
        self.K = K
        self.max_iter = max_iter
        self.random_states = random_states
        self.anim_data = anim_data

    def fit(self, X, bins=None):
        self.num_subs = len(X)
        self.X_list = X
        self.bins_list = bins
        self.subs = []
        for n in range(self.num_subs):
            random_state = None if self.random_states is None else self.random_states[n]
            proxy = EghMixture(self.K, self.max_iter, random_state=random_state, anim_data=self.anim_data)
            self.subs.append(proxy)

    def get_bins(self, n):
        return None if self.bins_list is None else self.bins_list[n]

    def get_sub(self, n):
        sub = self.subs[n]
        bins = self.get_bins(n)
        sub.fit(X=self.X_list[n], bins=bins)
        return sub

    def unified_fit(self, X, bins=None):
        assert bins is not None and len(bins) == 2

        self.fit(X, bins)

        for n, (sub, X) in enumerate(zip(self.subs, self.X_list)):
            print([n], X.shape)
            bins = self.get_bins(n)
            sub.initilize(X, bins=bins)

        for step in range(self.max_iter):
            params_list = []
            for sub in self.subs:
                sub._e_step()
                params_list.append(sub._m_step(step))

            new_params_list = self._u_step(params_list)
            for sub, params in zip(self.subs, new_params_list):
                sub.update_params(step, params)

    def _u_step(self, params_list):
        params_array = np.array(params_list)
        """
            [
                [(tR00, sigma00, tau00), (tR01, sigma01, tau01), ...],  # for UV ar 280
                [(tR10, sigma10, tau10), (tR11, sigma11, tau11), ...],  # for Xray at Q1
                ...
            ]
        """
        # print('params_array=', params_array)

        def obj_function(p):
            A, B = p
            return ( np.sum((A*params_array[1,:,0] + B - params_array[0,:,0])**2)
                    + np.sum(A*(params_array[1,:,1] - params_array[0,:,1])**2)
                    + np.sum(A*(params_array[1,:,2] - params_array[0,:,2])**2)
                    )

        A_init = self.bins_list[0]/self.bins_list[1]
        # print('A_init=', A_init)
        result = minimize(obj_function, (A_init, 0), args=())
        A, B = result.x

        A_ = 1/A
        B_ = -B/A

        tR_UV    = np.average([params_array[0,:,0], A*params_array[1,:,0] + B], axis=0)
        sigma_UV = np.average([params_array[0,:,1], A*params_array[1,:,1]], axis=0)
        tau_UV = np.average([params_array[0,:,2], A*params_array[1,:,2]], axis=0)

        tR_XR    = np.average([params_array[1,:,0], params_array[0,:,0]*A_ + B_], axis=0)
        sigma_XR = np.average([params_array[1,:,1], params_array[0,:,1]*A_], axis=0)
        tau_XR = np.average([params_array[1,:,2], params_array[0,:,2]*A_], axis=0)

        new_params_array = np.array([
                            np.array([tR_UV, sigma_UV, tau_UV]).T,
                            np.array([tR_XR, sigma_XR, tau_XR]).T,
                            ])
        # print('new_params_array=', new_params_array)
        return new_params_array



"""
    CurveSimilarity.py

        evaluation of simularity between curves

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize

USE_WLS_FOR_APPROX_MAPPING  = True
if USE_WLS_FOR_APPROX_MAPPING:
    import statsmodels.api  as sm
else:
    from scipy              import stats

BETTER_MAPPING_WANTED_SCORE = 0.5

def compute_chi_square(x, y, slope, intercept, weights):
    return np.sum( (x*slope + intercept - y)**2 * weights )/np.sum(weights)

class RegressInfo:
    def __init__(self, x, y, w, chisqr):
        self.x = x
        self.y = y
        self.w = w
        self.chisqr = chisqr

def compute_similarlity_factors(A, B, reg_info, a_curve, x_curve, reg_info_ref=None):
    size = len(x_curve.x)

    start = 0
    mapped_list = []
    max_peak_width = None
    for k, info in enumerate(x_curve.peak_info):
        left_i, top_i, right_i = info
        peak_width = right_i - left_i
        if max_peak_width is None or peak_width < max_peak_width:
            max_peak_width = peak_width
        if k < len(x_curve.boundaries):
            stop = x_curve.boundaries[k]
        else:
            stop = size
        i = np.arange(start, stop)
        j = A*i + B
        sync_y = a_curve.spline(j)
        top_j = A*top_i + B
        scale = x_curve.spline(top_i)/a_curve.spline(top_j)
        mapped_list.append(sync_y*scale)
        start = stop

    mapped_vector = np.hstack(mapped_list)
    spline_vector = x_curve.y_for_spline

    if False:
        import molass_legacy.KekLib.DebugPlot as plt
        fig = plt.figure()
        ax = fig.gca()
        ax.plot(spline_vector, label='spline_vector')
        b = x_curve.boundaries
        ax.plot(b, x_curve.spline(b), 'o', color='yellow', label='boundaries')
        ax.plot(mapped_vector, label='mapped_vector')
        ax.legend()
        fig.tight_layout()
        plt.show()

    if reg_info_ref is None:
        wpoints = reg_info.x
        reg_info_ref = reg_info
    else:
        wpoints = reg_info_ref.x

    weights = np.zeros(len(x_curve.x))
    hwidth = size*0.025
    for p in wpoints:
        f = max(0, int(p - hwidth))
        t = min(size, int(p + hwidth))
        weights[f:t] = 1

    y = np.array(reg_info_ref.y)
    x = np.array(reg_info_ref.x)
    y_ = A*x + B

    if False:
        fig = plt.figure(figsize=(21,6))
        fig.suptitle("compute_similarlity_factors")
        ax1 = fig.add_subplot(131)
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        ax3t = ax3.twinx()
        ax1.plot(a_curve.y, color='blue')
        ax1.plot(y, a_curve.spline(y), 'o', color='yellow')
        ax1.plot(y_, a_curve.spline(y_), 'o', color='orange')
        ax2.plot(x_curve.y, color='orange')
        ax2.plot(x, x_curve.spline(x), 'o', color='yellow')
        ax3.plot(mapped_vector, color='blue')
        ax3.plot(spline_vector, color='orange')
        ax3t.plot(weights)
        fig.tight_layout()
        plt.show()

    weights_ = weights/np.sum(weights)

    v_chisqr = np.sum(((mapped_vector - spline_vector)*weights_/x_curve.max_y)**2)
    # h_chisqr = reg_info.chisqr/max_peak_width**2
    w_ = reg_info_ref.w/np.sum(reg_info_ref.w)
    h_chisqr = np.sum(((y_ - x)*w_/max_peak_width)**2)
    # print('v_chisqr=', v_chisqr, 'h_chisqr=', h_chisqr)
    return v_chisqr, h_chisqr

class CurveSimilarity:
    def __init__(self, a_curve, x_curve, orig_a_curve=None, debug=False):
        if debug:
            from importlib import reload
            import Mapping.PeakMapper
            reload(Mapping.PeakMapper)
        from .PeakMapper import PeakMapper

        if orig_a_curve is None:
            orig_a_curve = a_curve
        self.orig_a_curve = orig_a_curve

        if False:
            # this is for plotting the untrimmed curves
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("CurveSimilarity debug (1)")
                ax1.plot(a_curve.x, a_curve.y)
                ax2.plot(x_curve.x, x_curve.y)
                fig.tight_layout()
                plt.show()

        pm = PeakMapper( a_curve, x_curve, debug=False )
        a_curve_, x_curve_ = pm.get_mapped_curves()
        self.a_curve    = a_curve_
        self.x_curve    = x_curve_
        slope, intercept = pm.best_info[1]

        if len(x_curve_.peak_info) == 1:
            score = self.compute_rough_mapping_score(slope, intercept)
            if score > BETTER_MAPPING_WANTED_SCORE:
                improved_result = self.try_improve_mapping(slope, intercept, score)
                if improved_result is not None:
                    if debug:
                        import molass_legacy.KekLib.DebugPlot as plt

                        def plot_mapped_curves(ax, a, b):
                            x = x_curve_.x
                            y = x_curve_.y
                            uv_x = x*a + b
                            uv_y = a_curve_.spline(uv_x)
                            ax.plot(x, y, color='orange')
                            ax.plot(x, uv_y*x_curve_.max_y/a_curve_.max_y, ":", color="blue")

                        with plt.Dp():
                            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                            fig.suptitle("CurveSimilarity debug (2)")
                            ax1.set_title("previous")
                            ax2.set_title("improved")
                            plot_mapped_curves(ax1, slope, intercept)
                            plot_mapped_curves(ax2, *improved_result)
                            fig.tight_layout()
                            plt.show()

                    slope, intercept = improved_result

        self.slope      = slope
        self.intercept  = intercept
        self.mapped_info = pm.mapped_info
        self.reliable = pm.reliable

    def get_mapped_info(self):
        return self.mapped_info

    def compute_rough_mapping_score(self, a, b):
        x = self.x_curve.x
        y = self.x_curve.y
        uv_x = x*a + b
        uv_y = self.a_curve.spline(uv_x)
        mapped_y = uv_y*self.x_curve.max_y/self.a_curve.max_y
        area1 = np.sum(y)
        area2 = np.sum(mapped_y)
        return abs(area1 - area2)*2/(area1 + area2)

    def try_improve_mapping(self, init_a, init_b, score):
        import logging
        logger = logging.getLogger(__name__)
        logger.info("tried to improve mapping due to the score %.3g", score)

        lower, _, upper = self.x_curve.peak_info[0]
        slice_ = slice(lower, upper)
        x_ = self.x_curve.x[slice_]
        y_ = self.x_curve.spline(x_)

        def objective(p):
            a, b = p
            uv_x_ = x_*a + b
            uv_y_ = self.a_curve.spline(uv_x_)
            mp_y_ = uv_y_*self.x_curve.max_y/self.a_curve.max_y
            return np.sum((mp_y_ - y_)**2)

        ret = minimize(objective, (init_a, init_b))
        new_score = self.compute_rough_mapping_score(*ret.x)

        if new_score < score:
            logger.info("got a better score %.3g", new_score)
            return ret.x
        else:
            logger.info("got a worse score %.3g, which will be discarded.", new_score)
            return None

    def mapped_value(self, i):
        j = self.slope * i + self.intercept
        return max(0, min(len(self.a_curve.y)-1, j))

    def mapped_int_value(self, i):
        j = int( self.slope * i + self.intercept + 0.5 )
        return max(0, min(len(self.a_curve.y)-1, j))

    def inverse_int_value(self, j, cut=True):
        i = int( (j - self.intercept)/self.slope + 0.5 )
        ret_i = max(0, min(len(self.x_curve.y)-1, i)) if cut else i
        return ret_i

    def plot_the_mapped_state(self, mapped_plot_closure=None):
        import molass_legacy.KekLib.DebugPlot as plt
        from molass_legacy.Elution.CurveUtils import simple_plot
        fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21,7))
        ax1, ax2, ax3 = axes
        simple_plot(ax1, self.a_curve, color='blue')
        simple_plot(ax2, self.x_curve, color='orange')
        simple_plot(ax3, self.x_curve, color='orange', legend=False)
        y = self.get_uniformly_mapped_a_curve()
        ax3.plot(y, ':', color='blue')
        if mapped_plot_closure is not None:
            mapped_plot_closure(ax3)
        fig.tight_layout()
        plt.show()

    def get_extended_x(self):
        start = self.inverse_int_value(0, cut=False)
        stop = self.inverse_int_value(len(self.a_curve.x), cut=False)
        x = np.arange(start, stop)
        return x

    def get_uniformly_mapped_a_curve(self, x=None):
        if x is None:
            x = self.x_curve.x
        j = self.slope * x + self.intercept
        scale = self.x_curve.max_y/self.orig_a_curve.max_y
        mapped_y = self.orig_a_curve.spline(j)*scale
        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Elution.CurveUtils import simple_plot
            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21,7))
            fig.suptitle("get_uniformly_mapped_a_curve")
            ax1, ax2, ax3 = axes
            simple_plot(ax1, self.orig_a_curve, "self.orig_a_curve")
            simple_plot(ax2, self.x_curve, "self.x_curve")
            simple_plot(ax3, self.x_curve, "self.x_curve", legend=False)
            ax3.plot(self.x_curve.x, mapped_y, label='mapped_y')
            ax3.legend()
            fig.tight_layout()
            plt.show()
            plt.pop()
        return mapped_y

    def compute_similarity_at( self, j, width_ratio=0.1, debug=False ):
        try:
            return self.compute_similarity_impl(j, width_ratio, debug=debug)
        except:
            # occured with 20170426
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            import logging
            etb = ExceptionTracebacker()
            logger = logging.getLogger(__name__)
            logger.warning("compute_similarity_at(%d) failed." % j + etb.last_lines() )
            return 0

    def compute_similarity_impl(self, j, width_ratio, debug=False):
        # print('compute_similarity_impl debug=', debug)

        ix  = (j - self.intercept)/self.slope
        x   = self.x_curve.x
        y   = self.x_curve.y

        i   = int(ix + 0.5)
        hw  = int( len(self.x_curve.x) * width_ratio/2 )
        start   = max( 0, i - hw )
        stop    = min( len(x), i + hw + 1 )

        cover_ratio = (stop - start)/(2*hw)     # =0.44 for right fc of HIF, which should not be discarded
        # print('cover_ratio=', cover_ratio)

        ratio_x = self.x_curve.spline(ix)/self.x_curve.max_y
        # print( 'ix=', ix, 'ratio_x=', ratio_x )

        slice_ = slice(start, stop)
        x_  = x[slice_]
        y_  = y[slice_]
        sy_ = self.x_curve.spline(x_)
        sy_min  = np.min(sy_)
        sy_max  = np.max(sy_)

        j_  = self.slope * x_ + self.intercept
        ay_ = self.a_curve.spline(j_)
        ay_min  = np.min(ay_)
        ay_max  = np.max(ay_)

        x_width = x_[-1] - x_[0]
        y_hight = sy_max - sy_min

        scale = y_hight/(ay_max - ay_min)
        ays = ay_ * scale
        x_spline = self.x_curve.spline

        def obj_func(params):
            A, B = params
            return np.sum((ays - (x_spline(x_ + A) + B ))**2)

        bounds = ((-x_width*0.3, x_width*0.3), (-y_hight*0.3, y_hight*0.3))

        result = minimize(obj_func, (0, 0), bounds=bounds)
        redchi = result.fun/(len(x_) - 2)
        diff_ratio  = np.sqrt(redchi) / y_hight
        similarity  = max( 0, 1 - diff_ratio )

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from matplotlib.patches     import Polygon
            from molass_legacy._MOLASS.SerialSettings         import get_setting

            opt_A, opt_B = result.x

            fy  = self.x_curve.spline( x_ + opt_A ) + opt_B

            ax  = plt.gca()
            ax.cla()
            ax.set_title( 'Curve Simualrity near j=%.3g in ' % j + get_setting('in_folder'), fontsize=16 )
            # ax.plot( x_, y_, color='orange' )
            ax.plot( x_, sy_, color='orange', label='original Xray (spline)' )
            ax.plot( x_, ays, label='mapped UV (spline)' )
            ax.plot( x_, fy, color='red', label='shifted Xray (spline)' )

            poly_points = list(zip( x_, ays )) + list( reversed( list( zip( x_, fy ) ) ) )
            diff_poly   = Polygon( poly_points, alpha=0.2 )
            ax.add_patch(diff_poly)
            ax.legend( fontsize=16 )

            ti  = i - start
            fc_y = ays[ti]
            ann_upper = fc_y  > (sy_min + sy_max)/2
            fc_y_delta = ( -1 if ann_upper else +1 ) * y_hight*0.2

            ax.annotate( "j=%.3g" % j, fontsize=16, xy=(i, fc_y), xytext=(i, fc_y + fc_y_delta ),
                        ha='center', va='center',
                        arrowprops=dict( headwidth=5, width=0.5, color='black', shrink=0.05) )

            tx  = x_[0] * 0.8  + x_[-1] * 0.2
            w   = 0.65 if ann_upper else 0.35
            ty  = sy_min * w + sy_max * (1 - w)

            ax.text( tx, ty, "Similarity=%.2g" % similarity, fontsize=50, alpha=0.2, va='center' )

            plt.tight_layout()
            plt.show()

        return similarity * cover_ratio
        # return 0

    def compute_whole_similarity(self):
        mapped_y = self.get_uniformly_mapped_a_curve()
        diff_area = np.sum(np.abs(mapped_y - self.x_curve.y))
        whole_area = self.x_curve.max_y * len(self.x_curve.y)
        return abs(1 -  diff_area/whole_area)



# coding: utf-8
"""

    ファイル名：   PilatusImageViewer.py

    処理内容：

       Pilatus 画像データの表示

    Copyright (c) 2015-2018, Masatsuyo Takahashi, KEK-PF
"""
from __future__ import division, print_function, unicode_literals
import os
import sys
import numpy as np
import pylab as pl
import copy

from molass_legacy.KekLib.OurTkinter         import Tk, ToolTip
from Preferences        import get_preference
from OurColorMaps       import CmapAlbulaLikeDynamic, Diverging
# from matplotlib.colors import SymLogNorm
# from PilatusColors      import LogNorm
from matplotlib.colors import LogNorm
import matplotlib.gridspec as gridspec
from ZoomPan            import ZoomPan
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backend_bases import key_press_handler
from matplotlib.figure  import Figure
from molass_legacy.KekLib.TkSupplements      import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib      import NavigationToolbar, CoordinateFormatter, DataCursor, ColorBar
from ControlKeyState    import set_shift_key_state, set_ctrl_key_state

value_shift = 10

class PilatusImageViewer( Tk.Toplevel ):
    def __init__( self, action, sample_id, exec_params ):
        Tk.Toplevel.__init__( self )
        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'

        # ここでアイコン設定すると、
        # 一瞬小さい Toplevel 画面が表示されるため、
        # 最後に設定する。
        # tk_set_icon_portable( self, 'synthesizer' )

        if action == 1:
            fig_title   = '%s (original)' % ( sample_id )
        else:
            fig_title   = '%s (adjusted)' % ( sample_id )

        self.title( fig_title )

        num_images = len( exec_params )

        fig = Figure( figsize=( num_images * 4, 5 ), dpi=100 )
        gs  = gridspec.GridSpec( 1, num_images )

        im_array_list   = []
        im_shift_list   = []
        ax_list         = []
        im_list         = []

        i = -1
        for param in exec_params:
            # print( '[%d] param=' % (i), param )

            i += 1
            file, im_array_orig = param[0:2]

            im_array = im_array_orig + value_shift
            im_array_list.append( im_array )
            if len( param ) >= 4:
                im_shift_list.append( param[2:4] )
            else:
                im_shift_list.append( [] )

            ax = fig.add_subplot( gs[i] )

            ax.set_title( file )

            if i ==0:
                numrows, numcols = im_array.shape
                self.formatter = CoordinateFormatter( numrows, numcols, im_array_list, value_shift )

            ax.format_coord = self.formatter
            ax.invert_yaxis()
            ax.axes.get_xaxis().set_visible(False)
            ax.axes.get_yaxis().set_visible(False)

            cmap_preference = get_preference( 'color_map' )
            if cmap_preference == 'Diverging':
                cmap_ = Diverging()
            else:
                vmin = np.min( im_array )
                vmax = np.max( im_array )
                v60  = np.percentile( im_array, 60 )
                logvmin, logvmax, logv60 = np.log(vmin), np.log(vmax), np.log( v60 )
                # cmap_mean = 0.2352 - 0.0590 * logvmin - 0.0393 * logvmax + 0.1480 * logv60
                cmap_mean = 1.1120 - 0.2301 * logvmin - 0.0162 * logvmax + 0.0176 * logv60
                # print( 'cmap_mean=%g' % ( cmap_mean ) )
                cmap_ = CmapAlbulaLikeDynamic( mean=cmap_mean )
            im = ax.imshow( im_array, cmap=cmap_, norm=LogNorm(), interpolation="nearest" )
            cb = ColorBar( im, ax )
            ax_list.append( ax )
            im_list.append( [ im, cmap_ ] )

        # fig.tight_layout()
        fig.set_tight_layout(True)

        canvas = FigureCanvasTkAgg( fig, master=self )

        if self.mplt_ge_2_2:
            canvas.draw()
        else:
            canvas.show()

        ToolTip( self, "<mouse-wheel> to zoom images; <drag> to move them.  <shift-click> to show pixel annotations; press <escape> to hide them." )

        scale = 2.0

        # Zoom 機能の設定
        for i in range( len(ax_list) ):
            # i 番目が先頭になるように並べ替えた ax_list を作る
            ax_list_copy = []
            ax_list_copy.append( ax_list[i] )
            for j in range( len(ax_list) ):
                if j != i:
                    ax_list_copy.append( ax_list[j] )
            # 並べ替えた ax_list_copy を使って Zoom 制御オブジェクトを作る
            zp      = ZoomPan( copy.copy(i), canvas, ax_list_copy[0], im_list[i] )
            figZoom = zp.zoom_factory( ax_list_copy, base_scale = scale )
            figPan  = zp.pan_factory( ax_list_copy )

        toolbar = NavigationToolbar( canvas, self )
        toolbar.update()

        canvas.get_tk_widget().pack( side=Tk.TOP, fill=Tk.BOTH, expand=1 )

        self.data_cursor = DataCursor( ax_list, action, value_shift )

        def on_key_press( event ):
            # print('you pressed %s' % event.key)

            # TODO:
            # key_press_handler(event, canvas, toolbar)

            if event.key == 'shift':
                set_shift_key_state( True )
            elif event.key == 'control':
                set_ctrl_key_state( True )

            self.data_cursor( 1, event, im_array_list, im_shift_list )

        def on_key_release( event ):
            if event.key == 'shift':
                set_shift_key_state( False )
            elif event.key == 'control':
                set_ctrl_key_state( False )

        def on_button_press( event ):
            self.data_cursor( 3, event, im_array_list, im_shift_list )

        canvas.mpl_connect( 'key_press_event',      on_key_press )
        canvas.mpl_connect( 'key_release_event',    on_key_release )
        canvas.mpl_connect( 'button_press_event',   on_button_press )

        def _quit():
            self.destroy()  # this is necessary on Windows to prevent
                            # Fatal Python Error: PyEval_RestoreThread: NULL tstate

        self.quit_button = Tk.Button(master=self, text='Quit', command=_quit)
        self.quit_button.pack(side=Tk.BOTTOM)

        tk_set_icon_portable( self, 'synthesizer' )

        self.update()   # 必要？

        return



"""
    StandardProcedure.py

    Copyright (c) 2020-2024, SAXS Team, KEK-PF
"""
import os
import logging
import copy
import numpy as np
from time import time
from molass_legacy.DataStructure.MeasuredData import MeasuredData
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.SerialAnalyzer.AbnormalityCheck import bubble_check
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from molass_legacy.Decomposer.DecompUtils import CorrectedBaseline
from molass_legacy.Models.ElutionCurveModels import EGHA, EMGA
from molass_legacy.Decomposer.DecompUtils import decompose_elution_better, make_range_info_impl
from molass_legacy.SerialAnalyzer.AnalyzerUtil import compute_conc_factor_util
from molass_legacy.Mapping.CurveSimilarity import CurveSimilarity

class MockEditor:
    def __init__(self):
        pass

    def update_range(self, *args):
        print('update_range:', *args)

class StandardProcedure:
    def __init__(self):
        self.logger = logging.getLogger()

    def load(self, folder, debug=False):
        self.folder = folder
        self.md = MeasuredData(folder)
        self.sd = None
        self.old_way = False
        if debug:
            self.md.plot()

    def load_old_way(self, folder):
        from molass_legacy.SerialAnalyzer.SerialDataLoader import SerialDataLoader
        from molass_legacy.SerialAnalyzer.SerialDataUtils import get_uv_filename
        self.folder = folder
        loader = SerialDataLoader()
        uv_file = get_uv_filename(folder)
        print("uv_file=", uv_file)
        if uv_file is None:
            set_setting('use_xray_conc', 1)
            loader.load_xray_data_only(folder)
        else:
            set_setting('use_xray_conc', 0)
            loader.load_from_folders(folder, uv_folder=folder, uv_file=uv_file )
        loader.wait_until_ready()
        self.sd = loader.get_current_object()
        print("sd loaded as %s in the old way" % type(self.sd))
        exclude = bubble_check(self.sd)
        if len(exclude) > 0:
            self.sd.exclude_intensities(exclude)
        self.old_way = True
        return self.sd

    def get_sd(self, whole=False, debug=False, set_offsets=True):
        from molass_legacy.SerialAnalyzer.SerialData import SerialData
        # datafiles, xray_array, uv_array, lvector, col_header, mtd_elution

        if self.old_way and self.sd is not None:
            return self.sd

        xr = self.md.xr

        if whole:
            i_slice = slice(0, None)
            j_slice = slice(0, None)
        else:
            i_slice = xr.i_slice
            j_slice = xr.j_slice
        datafiles = xr.files[j_slice]

        q = xr.vector[i_slice]
        shape = (len(datafiles), len(q), 3)
        xray_array = np.zeros(shape)
        xray_array[:,:,0] = np.array([ q for k in range(shape[0])])
        xray_array[:,:,1] = xr.data[i_slice, j_slice].T
        xray_array[:,:,2] = xr.error.data[i_slice, j_slice].T

        uv = self.md.uv
        if whole:
            i_slice = slice(None, None)
            j_slice = slice(None, None)
        else:
            i_slice = uv.i_slice
            j_slice = uv.j_slice
        uv_array = uv.data[i_slice, j_slice]
        lvector = uv.vector[i_slice]
        col_header = None   # implement if needed
        mtd_elution = None  # temp

        # restricted data
        data_info = copy.deepcopy([datafiles, xray_array, uv_array, lvector, col_header, mtd_elution])
        sd = SerialData(self.folder, self.folder, data_info=data_info)
        if set_offsets:
            sd.set_offsets_from_md(self.md)

        t0 = time()
        exclude = bubble_check(sd)
        print("took %.3g" % (time() - t0))
        if len(exclude) > 0:
            sd.exclude_intensities(exclude)

        if debug:
            md = MeasuredData(None, sd=sd)
            md.plot()

        self.sd = sd
        return sd

    def add_pre_recog(sp, sd):
        from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
        pre_recog = PreliminaryRecognition(sd)
        sd.pre_recog = pre_recog
        return pre_recog

    def get_curve_similarity(self):
        uv_curve = self.sd.get_uv_curve()
        xr_curve = self.sd.get_xray_curve()
        cs = CurveSimilarity(uv_curve, xr_curve)
        return cs

    def get_corrected_sd(self, proxy=True, debug=False):
        from molass_legacy.Mapping.MapperConstructor import create_mapper
        sd_orig = self.get_sd(debug=debug)
        if proxy:
            sd_orig.set_prerecog_proxy(self.md)
            pre_recog = sd_orig.pre_recog
            sd = sd_orig
        else:
            from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
            pre_recog = PreliminaryRecognition(sd_orig)
            sd = sd_orig._get_analysis_copy_impl(pre_recog)

        """
        approximate uv-xr mapping has been already done in MeasuredData.
        however, no baseline correction yet.
        """

        self.mapper = create_mapper(None, sd, sd_orig, pre_recog)
        sd.apply_baseline_correction(self.mapper.get_mapped_info())

        if debug:
            md = MeasuredData(None, sd=sd)
            md.plot()

        conc_factor = compute_conc_factor_util()
        sd.set_mc_vector(self.mapper, conc_factor)
        return sd

    def get_lrf_args(self, sd):
        from Extrapolation.PreviewData import PreviewData, PreviewOptions

        decomp_needed = self.get_decomp_necessity(sd)
        print('decomp_needed=', decomp_needed)
        if decomp_needed:
            use_elution_models = True
            decomp_info = self.get_decomp_info(sd)
            control_info = decomp_info.get_range_edit_info(logger=self.logger)
            """
            note that ranges are not shifted.
            consider refactoring which moves make_range_info_impl into ret
            """
            paired_ranges = make_range_info_impl(decomp_info.opt_recs_uv, control_info)
        else:
            use_elution_models = False
            decomp_info = None
            x_curve = sd.get_xray_curve()
            paired_ranges = x_curve.get_default_paired_ranges()
        conc_factor = compute_conc_factor_util()
        pdata = PreviewData(sd=sd, mapper=self.mapper, paired_ranges=paired_ranges,
                                decomp_info=decomp_info, conc_factor=conc_factor)
        popts = popts = PreviewOptions(use_elution_models=use_elution_models)
        return pdata, popts

    def get_preview_results(self, sd, gui=False, parent=None, return_ctrl=False, use_pool=True):
        pdata, popts = self.get_lrf_args(sd)

        editor = MockEditor()
        if use_pool:
            from LRF.LrfResultPool import LrfResultPool
            pool = LrfResultPool(pdata, popts)
            pool.run_solver(parent)
            if gui:
                pool.show_dialog(parent)
            return pool.get_better_results()
        else:
            from Extrapolation.PreviewController import PreviewController
            ctrl = PreviewController(dialog=parent, editor=editor)
            ctrl.run_solver(parent, pdata, popts)
            if gui:
                ctrl.show_dialog()

            if return_ctrl:
                return ctrl
            else:
                return ctrl.solver_results

    def get_decomp_necessity(self, sd):
        self.get_guide_info(sd)
        return self.guide_info.decomp_proc_needed()

    def get_guide_info(self, sd):
        from molass_legacy.QuickAnalysis.AnalysisGuide import get_analysis_guide_info
        self.discomfort = sd.compute_scattering_baseline_discomfort()
        self.guide_info = get_analysis_guide_info(self.mapper, None, self.discomfort)

    def get_decomp_info(self, sd):
        corbase_info = CorrectedBaseline(sd, self.mapper)
        min_score = None
        min_ret = None
        for k, model in enumerate([EMGA(), EGHA()]):
            ret = decompose_elution_better(corbase_info, self.mapper, model,
                                    logger=self.logger)
            score = ret.compute_synthetic_score()
            print([k], 'score=', score)
            if min_score is None or score < min_score:
                min_score = score
                min_ret = ret

        return min_ret



"""
    Optimizer.SecInspection.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Slider
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from SecTheory.BasicModels import robust_single_pore_pdf as monopore_pdf

class SecInspection(Dialog):
    def __init__(self, parent, js_canvas):
        self.parent = parent
        self.js_canvas = js_canvas
        dsets = js_canvas.dsets
        ecurve = dsets[0][0]
        self.x = ecurve.x
        self.y = ecurve.y

        fullopt = js_canvas.fullopt
        params = js_canvas.get_current_params()
        self.separate_params = fullopt.split_params_simple(params)
        self.xr_baseline = fullopt.xr_baseline(self.x, self.separate_params[1])
        Dialog.__init__(self, parent, "SEC Inspection", visible=False)

    def get_xr_params(self):
        return self.separate_params[0]

    def get_rg_params(self):
        return self.separate_params[2]

    def get_sec_params(self):
        if len(self.separate_params[-1]) == 2:
            sec_params = self.separate_params[-2]
        else:
            sec_params = self.separate_params[-1]
        return sec_params

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X)

        fig = plt.figure(figsize=(12,5))
        self.fig = fig

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        self.start_interactive()

        self.popup_menu = None
        # self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def start_interactive(self):
        fig = self.fig
        gs = GridSpec(1,4)
        ax1 = fig.add_subplot(gs[0,0:2])

        t0, rp, N, me, T, mp = self.get_sec_params()
        print("rp, t0, N, me, T, mp=", rp, t0, N, me, T, mp)

        recs = [
            ("$t_0$", -300, 300, t0),
            ("$r_p$", 30, 200, rp),
            ("N", 30, 3000, N),
            ("$m_e$", -0.1, 3.1, me),
            ("T", 0.1, 10, T),
            ("$m_p$", -0.1, 3.1, mp),
            ("$K=N \cdot T$", 100, 4000, N*T),
            ]

        x = self.x
        y = self.y

        ax1.plot(x, y, color="orange")

        xr_params = self.get_xr_params()
        rg_params = self.get_rg_params()
        rho = rg_params/rp
        rho[rho > 1] = 1

        curves = []
        for k, r in enumerate(rho):
            np_ = N * (1 - r)**me
            tp_ = T * (1 - r)**mp
            my = xr_params[k] * monopore_pdf(x - t0, np_, tp_)
            curve, = ax1.plot(x, my, ":")
            curves.append(curve)

        sliders = []

        self.synchronizing = False

        def update(i, val):
            if self.synchronizing:
                return

            t0, rp, N, me, T, mp, K = self.get_values()
            rho = rg_params/rp
            rho[rho > 1] = 1

            if i == 2:
                T = K/N
                self.after(100, lambda : self.sychrionize(N, T, K))
            elif i == 4:
                N = K/T
                self.after(100, lambda : self.sychrionize(N, T, K))
            elif i == 6:
                ratio = np.sqrt(K/(N*T))
                N = N*ratio
                T = T*ratio
                self.after(100, lambda : self.sychrionize(N, T, K))

            for k, r in enumerate(rho):
                np_ = N * (1 - r)**me
                tp_ = T * (1 - r)**mp
                my = xr_params[k] * monopore_pdf(x - t0, np_, tp_)
                curves[k].set_ydata(my)
            fig.canvas.draw_idle()

        for i, (label, valmin, valmax, valinit) in enumerate(recs):
            ax = fig.add_axes([0.6, 0.9 - 0.1*i, 0.3, 0.03])
            slider  = Slider(ax, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, i_=i: update(i_, val))
            sliders.append(slider)

        self.sliders = sliders

        fig.tight_layout()
        self.mpl_canvas.draw()
        self.sliders = sliders

    def get_values(self):
        return [slider.val for slider in self.sliders]

    def sychrionize(self, N, T, K):
        self.synchronizing = True
        self.sliders[2].set_val(N)
        self.sliders[4].set_val(T)
        self.sliders[6].set_val(K)
        self.after(300, self.rest_synchronizing)

    def rest_synchronizing(self):
        self.synchronizing = False

    def on_figure_click(self, event):
        if event.button == 3:
            self.show_popup_menu(event)
            return

    def show_popup_menu(self, event):
        from molass_legacy.KekLib.TkUtils import split_geometry

        self.create_popup_menu(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu(self, event):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0 )
            self.popup_menu.add_command(label='Draw Factor Curves', command=self.draw_factor_curves)

    def draw_factor_curves(self, debug=True):
        if debug:
            import Optimizer.FactorCurvePlot
            from importlib import reload
            reload(Optimizer.FactorCurvePlot)
        from molass_legacy.Optimizer.FactorCurvePlot import FactorCurvePlot

        print("draw_factor_curves")

        plot = FactorCurvePlot(self.parent, self)
        plot.show()



import os
import sys
import platform

thisdir_ = os.path.dirname( os.path.abspath( __file__ ) )
sys.path.append( thisdir_ )

def get_version_string():
    return 'KekLib 1.1.4 (2019-11-06 python %s %s)' % ( platform.python_version(), platform.architecture()[0] )



"""
    Models.Stochastic.DatablockUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

MIN_QUALITY = 0.3       # to correct under-estimated qualities as in 20210727/data01

def load_datablock_list(moment_rg_file, column_type):
    in_folders = []
    block_list = []
    record_list = None

    def break_procedure():
        nonlocal record_list
        if record_list is None:
            return

        block = np.array(record_list)
        too_low = block[:,3] < MIN_QUALITY
        block[too_low,3] = MIN_QUALITY
        block_list.append(block)
        record_list = None

    with open(moment_rg_file) as fh:
        skip = False
        for k, line in enumerate(fh):
            fields = line[:-1].split(',')
            if fields[1] == '0':
                if fields[3] == column_type:
                    print([k], line[:-1])
                    in_folders.append(fields[0])
                    break_procedure()
                    record_list = []
                    skip = False
                else:
                    skip = True
            else:
                if skip:
                    continue
                print([k], line[:-1])
                record_list.append(np.array([float(v) for v in fields[3:]]))
        break_procedure()

    for k, block in enumerate(block_list):
        print([k], block.shape, block.dtype)

    return in_folders, block_list

def get_block_lrf_src(batch, in_folder):
    from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings, set_setting
    clear_temporary_settings()      # required to avoid
    set_setting('in_folder', in_folder)
    sd = batch.load_data(in_folder)
    batch.prepare(sd, min_num_peaks=2)
    uv_x, uv_y, xr_x, xr_y, baselines = batch.get_curve_xy(return_baselines=True)
    uv_y_ = uv_y - baselines[0]
    xr_y_ = xr_y - baselines[1]
    uv_peaks, xr_peaks = batch.get_modeled_peaks(uv_x, uv_y_, xr_x, xr_y_, affine=True, min_area_prop=None)
    batch.set_lrf_src_args1(uv_x, uv_y, xr_x, xr_y, baselines)  # task: do not require users to do this
    lrf_src = batch.get_lrf_source()
    return lrf_src
()
def get_rg_curves(rg_folders_root, column_type):
    import os
    from RgProcess.RgCurveProxy import RgCurveProxy
    print("get_rg_curves: rg_folders_root=", rg_folders_root)
    mode = 0 if column_type == '1' else 1
    ret_curves = []
    for k, node in enumerate(os.listdir(rg_folders_root)):
        if k % 2 == mode:
            print([k], node)
            fg_folder = os.path.join(rg_folders_root, node)
            rg_curve = RgCurveProxy(None, fg_folder)
            ret_curves.append(rg_curve)
    return ret_curves


# coding: utf-8
"""
    DriftSimulation.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter         import Tk, Dialog
from molass_legacy.KekLib.TkUtils            import is_low_resolution
from CanvasFrame        import CanvasFrame
from DriftAnalyzer      import DriftAnalyzer
from DataModels         import _GuinierPorod

class DriftSimulation:
    def __init__( self, sd, mapper ):
        self.sd     = sd
        self.mapper = mapper
        self.data   = sd.intensity_array
        self.ecurve = mapper.x_curve

    def create_sim_data( self ):

        analyzer = DriftAnalyzer( self.sd, self.mapper )
        drift_params, gp_params = analyzer.solve()
        self.drift_params = drift_params

        if True:
            G_bg    = drift_params['Gbg']
            Rg_bg   = drift_params['Rg']
            d_bg    = drift_params['d']
            D_bg    = drift_params['slope']
            E_bg    = drift_params['intercept']
        else:
            # manually worked values for OA_Ald_Fer
            G_bg, Rg_bg, d_bg   = 0.793, 254, 3.5
            D_bg, E_bg = -0.000107, 0.779

        data    = self.data
        ecurve  = self.ecurve
        print( data.shape )
        num_elutions = data.shape[0]
        x   = data[0,:,0]

        y_sub_bg    = _GuinierPorod( G_bg, Rg_bg, d_bg, x )

        data_list = []
        base_list = []

        start   = 0
        for k, info in enumerate( ecurve.peak_info ):
            if k < len(ecurve.boundaries):
                stop    = ecurve.boundaries[k]
            else:
                stop    = num_elutions

            print( (start, stop) )
            p   = int(info[1]+0.5)
            top = ecurve.y[p]
            G, Rg, d    = gp_params[k]

            for j in range(start, stop):
                s   = ecurve.y[j]/top
                y   = _GuinierPorod( G*s, Rg, d, x )
                G_  = D_bg * j + E_bg
                y_drift = _GuinierPorod( G_, Rg_bg, d_bg, x )
                y_  = y + y_drift
                y_sub   = y_ - y_sub_bg
                data_list.append( np.vstack( [ x, y_sub, data[j,:,2 ] ] ).T )
                base_list.append( y_drift - y_sub_bg )

            if k > 0:
                print( ecurve.boundaries[k-1] )
            print( k, info )
            start = stop

        return np.array( data_list ), np.array( base_list )

def apply_simulated_baseline_correction( drift_params, data, progress_cb=None ):
    x   = data[0,:,0]

    print( 'drift_params=', drift_params )

    Gbg = drift_params['Gbg']
    Rg  = drift_params['Rg']
    d   = drift_params['d']
    slope   = drift_params['slope']
    intercept   = drift_params['intercept']

    num_elutions = data.shape[0]

    stain_bg    = _GuinierPorod( Gbg, Rg, d, x )

    for j in range( num_elutions ):
        G   = slope * j + intercept
        stain = _GuinierPorod( G, Rg, d, x )
        data[j,:,1] += stain_bg - stain
        if progress_cb is not None:
            progress_cb( j )

class BicomponentSimulator( Dialog ):
    def __init__( self, parent ):
        self.parent = parent
        self.x  = np.linspace( 0.003, 0.02, 50 )
        self.xx = self.x**2

    def show( self ):
        title   = "BicomponentSimulator"
        Dialog.__init__( self, self.parent, title )

    def body( self, body_frame ):   # overrides parent class method

        figsize = ( 18, 6 ) if is_low_resolution() else ( 21, 7 )
        self.canvas_frame = canvas_frame = CanvasFrame( body_frame, figsize=figsize )
        canvas_frame.pack()
        fig = canvas_frame.fig
        ax1  = fig.add_subplot( 131 )
        ax2  = fig.add_subplot( 132 )
        ax3  = fig.add_subplot( 133 )
        fig.tight_layout()

        self.draw( ax1, ax2, ax3 )

    def draw( self, ax1, ax2, ax3 ):

        ax1.set_title( "Guinier Plot of Protein with Stain" )

        yb  = _GuinierPorod( 0.01, 300, 3.5, self.x )
        y   = _GuinierPorod( 1.0, 35, 3, self.x )

        ax1.plot( self.xx, np.log( y + yb ),  label='protein + stain' )
        ax1.plot( self.xx, np.log( y ), label='protein'  )
        ax1.plot( self.xx, np.log( yb ), label='stain' )
        ax1.legend()

        ax2.set_title( "Guinier Plot of Protein with Stain" )
        ax2.plot( self.xx, np.log( y + yb ),  label='protein + stain' )
        ax2.plot( self.xx, np.log( y ), label='protein'  )
        ax2.legend()

        ax3.set_title( "Linear Plot of Protein with Stain" )
        ax3.plot( self.x, y + yb,  label='protein + stain' )
        ax3.plot( self.x, y, label='protein'  )
        ax3.legend()

        self.canvas_frame.show()



"""

    TrimmingInfo.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF
"""

TRIMMING_ITEMS = ["uv_restrict_list", "xr_restrict_list"]

class TrimmingInfo:
    def __init__(self, flag, start_, stop_, size_, extra=None):
        start = int(start_)     # int to avoid overspecification with numpy.int64
        stop = int(stop_)       # int to avoid overspecification with numpy.int64
        size = int(size_)       # int to avoid overspecification with numpy.int64
        self.flag   = flag
        self.start  = start
        self.end    = size - 1 if stop is None else stop - 1    # for plot (to be used)
        self.stop   = stop      # for slice
        self.size   = size
        self.extra  = extra
        self.items  = [flag, start, stop, size]

    def get_slice(self):
        return slice(self.start, self.stop)

    def get_safe_object(self):
        start_ = 0 if self.start is None else self.start
        stop_ = self.size if self.stop is None else self.stop
        return TrimmingInfo(self.flag, start_, stop_, self.size)

    def __repr__(self):
        return 'TrimmingInfo' + str( (self.flag, self.start, self.stop, self.size) )

    def __iter__(self):
        return iter([self.flag, self.start, self.stop, self.size])

    def __getitem__(self, item):
        return self.items[item]

def save_trimming_txt(path, info_list=None):
    if info_list is None:
        from molass_legacy._MOLASS.SerialSettings import get_setting
        info_list = [get_setting(item) for item in TRIMMING_ITEMS]
    fh = open(path, "w")
    for ti in info_list:
        fh.write(str(ti) + "\n")
    fh.close()

def load_trimming_txt(path):
    """
    trimming.txt example

    [TrimmingInfo(1, 656, 1107, 1600), None]
    [TrimmingInfo(1, 650, 1101, 1600), TrimmingInfo(1, 8, 1018, 1178)]

    """
    import numpy as np  # for eval
    fh = open(path)
    ret_list = []
    for line in fh:
        # print(line)
        line_list = eval(line, globals(), locals())
        ret_list.append(line_list)
    fh.close()
    return ret_list

def restore_trimming_info_impl(trimming_txt, logger=None):
    from molass_legacy._MOLASS.SerialSettings import set_setting

    if logger is None:
        import logging
        logger = logging.getLogger(__name__)

    ret_list = load_trimming_txt(trimming_txt)
    for item, info in zip(TRIMMING_ITEMS, ret_list):
        set_setting(item, info)
        logger.info("%s restored as %s", item, str(info))

def get_trimming_info_list():
    from molass_legacy._MOLASS.SerialSettings import get_setting
    info_list = []
    for item in TRIMMING_ITEMS:
        info_list.append(get_setting(item))
    return info_list

def set_trimming_info_list(info_list, logger=None):
    from molass_legacy._MOLASS.SerialSettings import set_setting

    if logger is None:
        import logging
        logger = logging.getLogger(__name__)

    for item, info in zip(TRIMMING_ITEMS, info_list):
        set_setting(item, info)
        logger.info("%s restored as %s", item, str(info))

def get_trimming_slices_from_list(info_list):
    ret_slices = []
    for list_ in info_list:
        if list_ is None:
            start, stop = None, None
        else:
            start = list_.start
            stop = list_.stop
        ret_slices.append(slice(start, stop))
    return ret_slices

def get_trimming_ends_from_list(info_list):
    if info_list is None:
        return [(0, -1), (0, -1)]

    ret_ends = []
    for list_ in info_list:
        if list_ is None:
            start, end = 0, -1
        else:
            start = list_.start
            end = list_.end
        ret_ends.append((start, end))
    return ret_ends

def seem_to_be_manually_trimmed():
    from molass_legacy._MOLASS.SerialSettings import get_setting

    manually_trimmed = get_setting("manually_trimmed")
    return manually_trimmed

def get_mapped_info(mapping, size, tinfo):
    A, B = mapping
    start = max(0, int(round(tinfo.start*A + B)))
    stop = min(size, int(round(tinfo.stop*A + B)))
    return TrimmingInfo(1, start, stop, size)

def get_wider_info(tinfo1, tinfo2):
    assert tinfo1.size == tinfo2.size
    start = min(tinfo1.start, tinfo2.start)
    stop = max(tinfo1.stop, tinfo2.stop)
    return TrimmingInfo(1, start, stop, tinfo1.size)



"""
    Optimizer.ParamsInspection.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk, Dialog

ENABLE_OPTIONAL_BUTTONS = False 

class ParamsInspection(Dialog):
    def __init__(self, parent, params, dsets, optimizer, state_info=None):
        self.parent = parent
        self.params = params
        self.dsets = dsets
        self.optimizer = optimizer
        self.state_info = state_info
        Dialog.__init__(self, parent, "Parameter Inspection", visible=False, location="lower right")

    def show(self):
        self._show()

    def body(self, body_frame):
        self.psheet = self.optimizer.params_type.get_params_sheet(body_frame, self.params, self.dsets, self.optimizer)
        self.psheet.enable_copy()
        self.psheet.pack()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, padx=50)

        num_buttons = 2 if self.state_info is None else 3
        if ENABLE_OPTIONAL_BUTTONS:
            num_buttons += 2
        for k in range(num_buttons):
            box.columnconfigure(k, weight=1)

        col = 0
        w = Tk.Button(box, text="◀ Close", width=10, command=self.cancel)
        w.grid(row=0, column=col, pady=10)

        if self.state_info is not None:
            w = Tk.Button(box, text="Highlight Active Parameters", width=24, command=self.highlight)
            col += 1
            w.grid(row=0, column=col, pady=10)

        w = Tk.Button(box, text="Save", width=10, command=self.save)
        col += 1
        w.grid(row=0, column=col, pady=10)

        if ENABLE_OPTIONAL_BUTTONS:
            w = Tk.Button(box, text="SEC Conformance", width=14, command=self.show_conformance)
            col += 1
            w.grid(row=0, column=col, pady=10)

            w = Tk.Button(box, text="SEC Simulation", width=14, command=self.show_simulator)
            col += 1
            w.grid(row=0, column=col, pady=10)

        self.bind("<Escape>", self.cancel)

    def highlight(self):
        try:
            self.psheet.highlight_active_params(self.state_info)
        except:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "highlight failed: ", n=10)
            MessageBox.showerror("Error", "Not supported yet.", parent=self.parent)

    def save(self):
        import os
        from tkinter import filedialog
        from molass_legacy._MOLASS.SerialSettings import get_setting

        initialdir = get_setting("optjob_folder")
        f = filedialog.asksaveasfilename(
            title="名前を付けて保存",
            # filetypes=[("Excel Books", "*.xlsx"), ("CSV Text Files", "*.csv")],
            filetypes=[("CSV Text Files", "*.csv")],
            initialdir=initialdir,
            initialfile="parameters.csv",
            parent=self,
            )
        if not f:
            return

        self.psheet.save_as(f)

    def show_simulator(self, debug=True):
        if debug:
            from importlib import reload
            import SecTheory.SecSimulator
            reload(SecTheory.SecSimulator)
        from SecTheory.SecSimulator import SecSimulator

        self.grab_set()  # temporary fix to the grab_release problem
        dialog = SecSimulator(self, self.optimizer, self.params)
        dialog.show()
        self.grab_set()  # temporary fix to the grab_release problem

    def show_conformance(self, debug=True):
        if debug:
            from importlib import reload
            import SecTheory.SecParamsPlot
            reload(SecTheory.SecParamsPlot)
        from SecTheory.SecParamsPlot import plot_sec_params

        lrf_info = self.optimizer.objective_func(self.params, return_lrf_info=True)
        self.grab_set()  # temporary fix to the grab_release problem
        plot_sec_params(self.optimizer, self.params, lrf_info)
        self.grab_set()  # temporary fix to the grab_release problem



"""
    XrSpecCurve.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""


class XrSpecCurve:
    def __init__(self, sd):
        D, E, qv, ecurve = sd.get_xr_data_separate_ly()
        self.x = qv
        j = ecurve.primary_peak_i
        self.y = D[:,j]

    def get_x(self):
        return self.x

    def get_y(self):
        return self.y



# coding: utf-8
"""
    ScatteringCurveViewer.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkUtils import adjusted_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable

class ScatteringCurveViewer( Dialog ):
    def __init__(self, parent, editor_frame, sd, ft, scaled_y):
        self.editor_frame = editor_frame
        self.uv_y = editor_frame.uv_y
        self.xr_y = scaled_y
        self.xr_j0 = sd.xr_j0
        self.qvector = sd.qvector
        self.data = sd.intensity_array[:,:,1].T
        self.ft = ft

        Dialog.__init__( self, parent, "Scattering Curve Viewer", visible=False )

    def show(self):
        self._show()

    def body( self, body_frame ):
        tk_set_icon_portable( self )

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14, 7))
        self.fig = fig
        self.axes = axes
        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.linear_plot()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X)

        upper_subbox = Tk.Frame(box)
        upper_subbox.pack(fill=Tk.X)
        lower_subbox = Tk.Frame(box)
        lower_subbox.pack()

        w = Tk.Button(upper_subbox, text="to Kratky", width=10, command=self.kratky_plot)
        w.pack(side=Tk.RIGHT, padx=5, pady=5)
        w = Tk.Button(upper_subbox, text="to Log", width=10, command=self.log_plot)
        w.pack(side=Tk.RIGHT, padx=5, pady=5)
        w = Tk.Button(upper_subbox, text="to Linear", width=10, command=self.linear_plot, default=Tk.ACTIVE)
        w.pack(side=Tk.RIGHT, padx=5, pady=5)

        w = Tk.Button(lower_subbox, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(lower_subbox, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def kratky_plot(self):
        self.plain_plot(kratky=True)

    def log_plot(self):
        self.plain_plot(log=True)

    def linear_plot(self):
        self.plain_plot()

    def plain_plot(self, log=False, kratky=False):
        for ax in self.axes:
            ax.cla()
            ax.set_xlabel('Q')
            if kratky:
                ax.set_ylabel('I/C*$Q^2$')
            else:
                if log:
                    ax.set_yscale('log')
                    ax.set_ylabel('log(I/C)')
                else:
                    ax.set_ylabel('I/C')

        fig = self.fig
        ax1, ax2 = self.axes
        x = self.qvector
        if kratky:
            xx = x*x
        ft = self.ft
        ft_ = [j - self.xr_j0 for j in ft]
        ft_label = r'(No. %d$\sim$%d)' % tuple(ft)

        plot_type = 'Kratky' if kratky else ('Log' if log else 'Linear')
        fig.suptitle(plot_type + " Plot of Normalized Curves", fontsize=20)
        ax1.set_title("Mapped UV Normalized " + ft_label, fontsize=16)
        ax2.set_title("Scaled Xray Normalized " + ft_label, fontsize=16)

        for j in range(ft_[0], ft_[1]+1):
            y = self.data[:,j]
            if kratky:
                ax1.plot(x, y/self.uv_y[j]*xx)
                ax2.plot(x, y/self.xr_y[j]*xx)
            else:
                ax1.plot(x, y/self.uv_y[j])
                ax2.plot(x, y/self.xr_y[j])

        fig.tight_layout()
        fig.subplots_adjust(top=0.85)
        self.mpl_canvas.draw()



"""
    SimpleDecompose.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import minimize
from scipy.spatial import distance
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Characteristic import CfSpace
from molass_legacy.Decomposer.FitRecord import FitRecord
from molass_legacy.Decomposer.ModelEvaluator import ModelEvaluator
from molass_legacy.Decomposer.UnifiedDecompResult import UnifiedDecompResult
from Selective.PeakProxy import PeakProxy
from molass_legacy.Models.CfsEvalPeaks import CFSE_WEIGHT
from Distance.JensenShannon import deformed_jsd

class NotImplementedError(Exception): pass

IGNORABLE_PROPORTION = 0.01
PENALTY_SCALE = 1000

def decompose_curve(x, arg_y, model, num_peaks, props=None, using_hybrid=False, using_cfs=False, logger=None, debug=False):
    y = arg_y.copy()
    y[y < 0] = 0        # negative values are not allowed in distance.jensenshannon()

    model_params = model.guess_multiple(x, y, num_peaks, debug=debug)
    num_params = len(model_params)      # num_params can be less than num_peaks in cases like pH6

    b = np.percentile(y, 95)

    if using_hybrid or using_cfs:
        cfs = CfSpace()
        cft0 = cfs.compute_cf(x, y)

    if props is not None:
        props = np.asarray(props)

    def objective(params, debug=False, debug_title=None):
        cy_list = []
        if props is not None:
            areas = []
        for p in params.reshape(model_params.shape):
            cy = model.func(x, *p)
            cy_list.append(cy)
            if props is not None:
                areas.append(np.sum(cy))
        ty = np.sum(cy_list, axis=0)

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(debug_title)
                ax.plot(x, y)
                for cy in cy_list:
                    ax.plot(x, cy, ":")
                ax.plot(x, ty, ":", color="red", lw=3)
                fig.tight_layout()
                plt.show()

        if using_hybrid:
            cft1 = cfs.compute_cf(x, ty)
            return np.log(np.sum((ty - y)**2)) + CFSE_WEIGHT*np.log(np.sum((cft1 - cft0)**2))
            # return np.log(np.sum((ty - y)**2)) + CFSE_WEIGHT*np.log(np.sum(np.abs(cft1 - cft0)**2))
        elif using_cfs:
            cft1 = cfs.compute_cf(x, ty)
            return np.sum((cft1 - cft0)**2)
            # return np.sum(np.abs(cft1 - cft0)**2)
        else:
            # fv = deformed_jsd(ty, y, b=b)
            ydiv = np.sum((ty - y)**2)
            if props is None:
                fv = ydiv
            else:
                props_ = np.array(areas)/np.sum(areas)
                pdiv = np.sum((props_ - props)**2)
                fv = np.log10(ydiv)*0.8 + np.log10(pdiv)*0.2
                if debug:
                    print("ydiv, pdiv=", ydiv, pdiv)

            if np.isnan(fv):
                # as in AhRR with EDM
                fv = np.inf
            return fv

    params = model_params.flatten()
    if debug:
        objective(params, debug=True, debug_title="before minimize")

    ret_edm = minimize(objective, params)
    if debug:
        objective(ret_edm.x, debug=True, debug_title="after minimize")

    opt_recs = []
    kno = 0
    area = np.sum(y)
    for k, params in enumerate(ret_edm.x.reshape(model_params.shape)):
        cy = model.func(x, *params)       # make it less expensive
        j = np.argmax(cy)
        peak = PeakProxy(top_x=x[j], top_y=cy[j])
        carea = np.sum(cy)
        area_prop = carea/area
        if area_prop > IGNORABLE_PROPORTION:
            fit_rec = FitRecord(kno, ModelEvaluator(model, params, sign=1), 0, peak, area=carea)
            opt_recs.append(fit_rec)
            kno += 1
        else:
            logger.info("element with area_prop=%g <= %g has been ingored in decompose_curve.", area_prop, IGNORABLE_PROPORTION)

    return opt_recs

def remove_the_smallest(x, opt_recs, recsname, logger):
    heights = []
    for rec in opt_recs:
        y = rec.evaluator(rec.peak.top_x)
        heights.append(y)
    i = np.argmin(heights)
    opt_recs.pop(i)

    logger.warning("removed the smallest %d-th component from %s, where heights=%s", i, recsname, str(heights))

    return opt_recs

def plot_both_results(x, y, uv_y, opt_recs, opt_recs_uv, title):

    def plot_results(ax, x, y, recs):
        ax.plot(x, y)
        cy_list = []
        for rec in recs:
            cy = rec.evaluator(x)
            ax.plot(x, cy, ":")
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        if len(ty) == len(x):
            # temporary fix for ValueError: x and y must have same first dimension, but have shapes (926,) and (1,)
            ax.plot(x, ty, ":", color="red", lw=3)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle(title)
        ax1.set_title("UV")
        ax2.set_title("XR")
        plot_results(ax1, x, uv_y, opt_recs_uv)
        plot_results(ax2, x, y, opt_recs)
        fig.tight_layout()
        ret = plt.show()

    return ret

def sort_and_renumber(opt_recs):
    opt_recs = sorted(opt_recs, key=lambda rec: rec.peak.top_x)
    for k, rec in enumerate(opt_recs):
        rec.kno = k
    return opt_recs

def decompose_elution_simply(x, y, uv_y, model, traditional_info, props=None, using_cfs=False, debug=False):

    logger = logging.getLogger(__name__)

    num_peaks = traditional_info.num_peaks
    mapper = traditional_info.mapper

    opt_recs = decompose_curve(x, y, model, num_peaks, props=props, using_cfs=using_cfs, logger=logger, debug=debug)
    opt_recs_uv = decompose_curve(x, uv_y, model, num_peaks, props=props, using_cfs=using_cfs, logger=logger, debug=debug)

    if debug:
        from molass_legacy.Decomposer.OptRecsUtils import compute_area_proportions
        uv_props = compute_area_proportions(x, opt_recs_uv)
        print("uv_props=", uv_props)
        xr_props = compute_area_proportions(x, opt_recs)
        print("xr_props=", xr_props)
        ret = plot_both_results(x, y, uv_y, opt_recs, opt_recs_uv, "decomposed")
        if not ret:
            return

    if abs(len(opt_recs) - len(opt_recs_uv)) > 0:
        if len(opt_recs) < len(opt_recs_uv):
            while len(opt_recs) < len(opt_recs_uv):
                opt_recs_uv = remove_the_smallest(x, opt_recs_uv, "opt_recs_uv", logger)
        else:
            while len(opt_recs) > len(opt_recs_uv):
                opt_recs = remove_the_smallest(x, opt_recs, "opt_recs", logger)

    if len(opt_recs) != len(opt_recs_uv):
        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("decompose_elution_simply debug")
                ax1.plot(x, uv_y)
                for rec in opt_recs_uv:
                    cy = rec.evaluator(x)
                    ax1.plot(x, cy, ":")
                ax2.plot(x, y)
                for rec in opt_recs:
                    cy = rec.evaluator(x)
                    ax2.plot(x, cy, ":")
                fig.tight_layout()
                plt.show()

        raise NotImplementedError("len(opt_recs) %d != len(opt_recs_uv) %d" % (len(opt_recs), len(opt_recs_uv)))

    opt_recs = sort_and_renumber(opt_recs)
    opt_recs_uv = sort_and_renumber(opt_recs_uv)

    if debug:
        from molass_legacy.Decomposer.OptRecsUtils import compute_area_proportions
        uv_props = compute_area_proportions(x, opt_recs_uv)
        print("uv_props=", uv_props)
        xr_props = compute_area_proportions(x, opt_recs)
        print("xr_props=", xr_props)
        plot_both_results(x, y, uv_y, opt_recs, opt_recs_uv, "sorted")

    simple_result = UnifiedDecompResult(
                    xray_to_uv=None,
                    x_curve=mapper.x_curve, x=x, y=y,
                    opt_recs=opt_recs,
                    max_y_xray = np.max(y),     # get this from traditional_info
                    model_name=model.get_name(),
                    decomposer=None,            # what is this used for?
                    uv_y=uv_y,
                    opt_recs_uv=opt_recs_uv,
                    max_y_uv = np.max(uv_y),    # get this from traditional_info
                    )

    simple_result.set_area_proportions()
    simple_result.remove_unwanted_elements()

    return simple_result



"""
    MplUtils.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.colors import BoundaryNorm as ParentNorm

class BoundaryNorm(ParentNorm):
    def __init__(self, *args, **kwargs):
        self.inv_ret = kwargs.pop("inv_ret", -9999)     # not sure for -9999, which seems ok for CorMap
        ParentNorm.__init__(self, *args, **kwargs)

    def inverse(self, value):
        try:
            return ParentNorm.inverse(self, value)
        except ValueError:
            # ignore because it is known as below.
            # BoundaryNorm is not invertible, so calling this method will always raise an error
            # users should set inv_ret appropriately
            pass

        return self.inv_ret
